import functools
import json
import uuid

from fief_client import FiefTokenResponse, FiefUserInfo
from pydantic import UUID4
from sqlalchemy import String, Text
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import MainBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class AdminSessionToken(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "admin_session_tokens"

    token: Mapped[str] = mapped_column(String(length=255), unique=True, nullable=False)
    raw_tokens: Mapped[str] = mapped_column(Text, nullable=False)
    raw_userinfo: Mapped[str] = mapped_column(Text, nullable=False)

    def __repr__(self) -> str:
        return f"AdminSessionToken(id={self.id})"

    @functools.cached_property
    def tokens(self) -> FiefTokenResponse:
        return json.loads(self.raw_tokens)

    @functools.cached_property
    def userinfo(self) -> FiefUserInfo:
        return json.loads(self.raw_userinfo)

    @functools.cached_property
    def user_id(self) -> UUID4:
        return uuid.UUID(self.userinfo["sub"])



import re

from wtforms import validators

from fief.services.localhost import is_localhost


class NotHTTPSURLValidationError(validators.ValidationError):
    def __init__(self):
        super().__init__("An HTTPS URL is required.")


class RedirectURLValidator(validators.Regexp):
    def __init__(self, message=None):
        regex = (
            r"^(?P<scheme>[a-z]+)://"
            r"(?P<host>[^\/\?:]+)"
            r"(?P<port>:[0-9]+)?"
            r"(?P<path>\/.*?)?"
            r"(?P<query>\?.*)?$"
        )
        super().__init__(regex, re.IGNORECASE, message)

    def __call__(self, form, field):
        message = self.message
        if message is None:
            message = "Invalid URL."

        match = super().__call__(form, field, message)
        scheme = match.group("scheme")
        host = match.group("host")

        if scheme == "http" and not is_localhost(host):
            raise NotHTTPSURLValidationError()



from fastapi import Request

from fief.apps.auth.forms.auth import LoginForm
from fief.apps.auth.forms.register import RegisterFormBase, get_register_form_class
from fief.apps.auth.forms.reset import ForgotPasswordForm, ResetPasswordForm
from fief.models import Tenant, Theme
from fief.repositories import OAuthProviderRepository, UserFieldRepository
from fief.templates import templates


class ThemePreview:
    def __init__(
        self,
        oauth_provider_repository: OAuthProviderRepository,
        user_field_repository: UserFieldRepository,
    ) -> None:
        self.oauth_provider_repository = oauth_provider_repository
        self.user_field_repository = user_field_repository

    async def preview(
        self, page: str, theme: Theme, *, tenant: Tenant, request: Request
    ) -> str:
        page_functions = {
            "login": self.preview_login,
            "register": self.preview_register,
            "forgot_password": self.preview_forgot_password,
            "reset_password": self.preview_reset_password,
        }
        return await page_functions[page](theme, tenant=tenant, request=request)

    async def preview_login(
        self, theme: Theme, *, tenant: Tenant, request: Request
    ) -> str:
        oauth_providers = await self.oauth_provider_repository.all()
        form = LoginForm(meta={"request": request, "csrf": False})
        context = {
            "request": request,
            "form": form,
            "oauth_providers": oauth_providers,
            "tenant": tenant,
            "theme": theme,
        }
        template = templates.get_template("auth/login.html")
        return template.render(context)

    async def preview_register(
        self, theme: Theme, *, tenant: Tenant, request: Request
    ) -> str:
        registration_user_fields = (
            await self.user_field_repository.get_registration_fields()
        )
        register_form_class: type[RegisterFormBase] = await get_register_form_class(
            registration_user_fields, None
        )
        oauth_providers = await self.oauth_provider_repository.all()
        form = register_form_class(meta={"request": request, "csrf": False})
        context = {
            "request": request,
            "form": form,
            "finalize": False,
            "oauth_providers": oauth_providers,
            "tenant": tenant,
            "theme": theme,
        }
        template = templates.get_template("auth/register.html")
        return template.render(context)

    async def preview_forgot_password(
        self, theme: Theme, *, tenant: Tenant, request: Request
    ) -> str:
        form = ForgotPasswordForm(meta={"request": request, "csrf": False})
        context = {
            "request": request,
            "form": form,
            "tenant": tenant,
            "theme": theme,
        }
        template = templates.get_template("auth/forgot_password.html")
        return template.render(context)

    async def preview_reset_password(
        self, theme: Theme, *, tenant: Tenant, request: Request
    ) -> str:
        form = ResetPasswordForm(meta={"request": request, "csrf": False})
        context = {
            "request": request,
            "form": form,
            "tenant": tenant,
            "theme": theme,
        }
        template = templates.get_template("auth/reset_password.html")
        return template.render(context)



from wtforms import BooleanField, StringField, validators

from fief.forms import ComboboxSelectMultipleField, CSRFBaseForm


class BaseRoleForm(CSRFBaseForm):
    name = StringField("Name", validators=[validators.InputRequired()])
    granted_by_default = BooleanField(
        "Granted by default",
        description="When enabled, the role will automatically be assigned to users when they register.",
    )
    permissions = ComboboxSelectMultipleField(
        "Permissions",
        query_endpoint_path="/admin/access-control/permissions/",
        label_attr="codename",
        validators=[validators.InputRequired()],
        choices=[],
        validate_choice=False,
    )


class RoleCreateForm(BaseRoleForm):
    pass


class RoleUpdateForm(BaseRoleForm):
    pass



from fastapi import APIRouter, Depends, HTTPException, Response, status

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.user_field import (
    get_paginated_user_fields,
    get_user_field_by_id_or_404,
    get_validated_user_field_create,
    get_validated_user_field_update,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, UserField
from fief.repositories import UserFieldRepository
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import (
    UserFieldCreated,
    UserFieldDeleted,
    UserFieldUpdated,
)

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/",
    name="user_fields:list",
    response_model=PaginatedResults[schemas.user_field.UserField],
)
async def list_user_fields(
    paginated_user_fields: PaginatedObjects[UserField] = Depends(
        get_paginated_user_fields
    ),
) -> PaginatedResults[schemas.user_field.UserField]:
    user_fields, count = paginated_user_fields
    return PaginatedResults(
        count=count,
        results=[
            schemas.user_field.UserField.from_orm(user_field)
            for user_field in user_fields
        ],
    )


@router.get(
    "/{id:uuid}", name="user_fields:get", response_model=schemas.user_field.UserField
)
async def get_user_field(
    user_field: UserField = Depends(get_user_field_by_id_or_404),
) -> UserField:
    return user_field


@router.post(
    "/",
    name="user_fields:create",
    response_model=schemas.user_field.UserField,
    status_code=status.HTTP_201_CREATED,
)
async def create_user_field(
    user_field_create: schemas.user_field.UserFieldCreate = Depends(
        get_validated_user_field_create
    ),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.user_field.UserField:
    existing_user_field = await repository.get_by_slug(user_field_create.slug)
    if existing_user_field is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_FIELD_SLUG_ALREADY_EXISTS,
        )

    user_field = UserField(**user_field_create.dict())
    user_field = await repository.create(user_field)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, user_field)
    trigger_webhooks(UserFieldCreated, user_field, schemas.user_field.UserField)

    return schemas.user_field.UserField.from_orm(user_field)


@router.patch(
    "/{id:uuid}", name="user_fields:update", response_model=schemas.user_field.UserField
)
async def update_user_field(
    user_field_update: schemas.user_field.UserFieldUpdate = Depends(
        get_validated_user_field_update
    ),
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.user_field.UserField:
    updated_slug = user_field_update.slug
    if updated_slug is not None and updated_slug != user_field.slug:
        existing_user_field = await repository.get_by_slug(updated_slug)
        if existing_user_field is not None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.USER_FIELD_SLUG_ALREADY_EXISTS,
            )

    user_field_update_dict = user_field_update.dict(exclude_unset=True)
    for field, value in user_field_update_dict.items():
        setattr(user_field, field, value)

    await repository.update(user_field)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, user_field)
    trigger_webhooks(UserFieldUpdated, user_field, schemas.user_field.UserField)

    return schemas.user_field.UserField.from_orm(user_field)


@router.delete(
    "/{id:uuid}",
    name="user_fields:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_user_field(
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(user_field)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, user_field)
    trigger_webhooks(UserFieldDeleted, user_field, schemas.user_field.UserField)



from fastapi import Cookie, Depends

from fief.crypto.token import get_token_hash
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import SessionToken
from fief.repositories import SessionTokenRepository
from fief.settings import settings


async def get_session_token(
    token: str | None = Cookie(None, alias=settings.session_cookie_name),
    repository: SessionTokenRepository = Depends(
        get_workspace_repository(SessionTokenRepository)
    ),
) -> SessionToken | None:
    if token is not None:
        token_hash = get_token_hash(token)
        return await repository.get_by_token(token_hash)
    return None



from typing import Any, Type

from fastapi import Request
from wtforms import (
    BooleanField,
    FieldList,
    Form,
    FormField,
    IntegerField,
    SelectField,
    StringField,
    validators,
)
from wtforms.utils import unset_value

from fief.forms import CSRFBaseForm, TimezoneField
from fief.models import UserField, UserFieldType


class FormFieldTuple(FormField):
    def process(self, formdata, data=unset_value, extra_filters=None):
        """
        Transform input data in form of a tuple to a dict labelled by field key.

        The trick is to iterate over the form class to detect the field definitions
        (the logic is borrowed form FormMeta) and assign them the tuple value
        in the same order.
        """
        data_dict = {}
        if data:
            data_list = list(data)
            for name in dir(self.form_class):
                if not name.startswith("_"):
                    unbound_field = getattr(self.form_class, name)
                    if hasattr(unbound_field, "_formfield"):
                        data_dict[name] = data_list.pop(0)

        return super().process(formdata, data_dict, extra_filters)

    def populate_obj(self, obj, name):
        """
        Transform data labelled by field key into a tuple, in the order of the fields.
        """
        data_list = []
        for field in self.form:
            data_list.append(field.data)
        setattr(obj, name, tuple(data_list))


class SettableDict(dict):
    def __setattr__(self, __name: str, __value: Any) -> None:
        self[__name] = __value


class FormFieldPopulateJSON(FormField):
    def populate_obj(self, obj, name):
        d = SettableDict()
        self.form.populate_obj(d)
        setattr(obj, name, d)


class UserFieldConfigurationBase(Form):
    at_registration = BooleanField("Ask at registration")
    at_update = BooleanField("Ask at profile update")
    required = BooleanField("Required")

    def set_dynamic_parameters(self):
        pass


class UserFieldConfigurationString(UserFieldConfigurationBase):
    default = StringField("Default value")


class UserFieldConfigurationInteger(UserFieldConfigurationBase):
    default = IntegerField("Default value")


class UserFieldConfigurationBoolean(UserFieldConfigurationBase):
    default = BooleanField("Enabled")


class UserFieldConfigurationDate(UserFieldConfigurationBase):
    pass


class UserFieldConfigurationDateTime(UserFieldConfigurationBase):
    pass


class UserFieldConfigurationChoiceItem(Form):
    value = StringField(render_kw={"placeholder": "Value"})
    label = StringField(render_kw={"placeholder": "Label"})


class UserFieldConfigurationChoice(UserFieldConfigurationBase):
    choices = FieldList(
        FormFieldTuple(UserFieldConfigurationChoiceItem),
        "Choices",
        min_entries=1,
    )
    default = SelectField("Default value")

    def set_dynamic_parameters(self):
        choices = self.data["choices"]
        if choices is not None:
            self.default.choices = [("", "")] + [
                (choice["value"], choice["label"])
                for choice in choices
                if choice["value"] is not None
            ]


class UserFieldConfigurationPhoneNumber(UserFieldConfigurationBase):
    pass


class UserFieldConfigurationAddress(UserFieldConfigurationBase):
    pass


class UserFieldConfigurationTimezone(UserFieldConfigurationBase):
    default = TimezoneField("Default value")


CONFIGURATION_FORM_CLASS_MAP: dict[UserFieldType, type[Form]] = {
    UserFieldType.STRING: UserFieldConfigurationString,
    UserFieldType.INTEGER: UserFieldConfigurationInteger,
    UserFieldType.BOOLEAN: UserFieldConfigurationBoolean,
    UserFieldType.DATE: UserFieldConfigurationDate,
    UserFieldType.DATETIME: UserFieldConfigurationDateTime,
    UserFieldType.CHOICE: UserFieldConfigurationChoice,
    UserFieldType.PHONE_NUMBER: UserFieldConfigurationPhoneNumber,
    UserFieldType.ADDRESS: UserFieldConfigurationAddress,
    UserFieldType.TIMEZONE: UserFieldConfigurationTimezone,
}


class BaseUserFieldForm(CSRFBaseForm):
    name = StringField("Name", validators=[validators.InputRequired()])
    slug = StringField("Slug", validators=[validators.InputRequired()])


class UserFieldCreateForm(BaseUserFieldForm):
    type = SelectField(
        "Type",
        choices=UserFieldType.choices(),
        coerce=UserFieldType.coerce,
        default=UserFieldType.STRING.value,
        validators=[validators.InputRequired()],
    )

    @classmethod
    async def get_form_class(cls, request: Request) -> Type["UserFieldCreateForm"]:
        formdata = None
        if request.method in {"POST", "PUT", "PATCH"}:
            formdata = await request.form()
        base_form = cls(formdata, meta={"request": request})
        field_type = base_form.data["type"]
        if field_type is not None and field_type in CONFIGURATION_FORM_CLASS_MAP:
            configuration_form_class = CONFIGURATION_FORM_CLASS_MAP[field_type]

            class UserFieldForm(UserFieldCreateForm):
                configuration = FormFieldPopulateJSON(configuration_form_class)

            return UserFieldForm
        return cls


class UserFieldUpdateForm(BaseUserFieldForm):
    @classmethod
    async def get_form_class(cls, user_field: UserField) -> Type["UserFieldUpdateForm"]:
        class UserFieldForm(UserFieldUpdateForm):
            configuration = FormFieldPopulateJSON(
                CONFIGURATION_FORM_CLASS_MAP[user_field.type]
            )

        return UserFieldForm



from enum import StrEnum


class EmailTemplateType(StrEnum):
    BASE = "BASE"
    WELCOME = "WELCOME"
    FORGOT_PASSWORD = "FORGOT_PASSWORD"

    def get_display_name(self) -> str:
        display_names = {
            EmailTemplateType.BASE: "Base",
            EmailTemplateType.WELCOME: "Welcome",
            EmailTemplateType.FORGOT_PASSWORD: "Forgot password",
        }
        return display_names[self]



import urllib.parse
from typing import Any, Literal, TypedDict

from fastapi import Depends, Header, Query, Request
from fief_client import FiefUserInfo

from fief.dependencies.admin_session import get_userinfo
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.pagination import (
    Ordering,
    OrderingGetter,
    Pagination,
    get_pagination,
)
from fief.dependencies.workspace import get_admin_user_workspaces
from fief.models import Workspace


async def get_layout(hx_request: bool = Header(False)) -> str:
    if hx_request:
        return "admin/layout_boost.html"
    return "admin/layout.html"


class BaseContext(TypedDict):
    request: Request
    layout: str
    hx_target: str | None
    user: FiefUserInfo
    current_workspace: Workspace
    workspaces: list[Workspace]


async def get_base_context(
    request: Request,
    hx_target: str | None = Header(None),
    layout: str = Depends(get_layout),
    userinfo: FiefUserInfo = Depends(get_userinfo),
    current_workspace: Workspace = Depends(get_current_workspace),
    workspaces: list[Workspace] = Depends(get_admin_user_workspaces),
) -> BaseContext:
    return {
        "request": request,
        "layout": layout,
        "hx_target": hx_target,
        "user": userinfo,
        "current_workspace": current_workspace,
        "workspaces": workspaces,
    }


class DatatableColumn:
    def __init__(
        self,
        title: str,
        slug: str,
        renderer_macro: str,
        *,
        ordering: str | None = None,
        renderer_macro_kwargs: dict[str, Any] | None = None,
    ) -> None:
        self.title = title
        self.slug = slug
        self.ordering = ordering
        self.renderer_macro = renderer_macro
        self.renderer_macro_kwargs = (
            renderer_macro_kwargs if renderer_macro_kwargs is not None else {}
        )


class DatatableQueryParameters:
    def __init__(
        self,
        pagination: Pagination,
        ordering: Ordering,
        columns: list[str],
    ) -> None:
        limit, skip = pagination
        self.limit = limit
        self.skip = skip
        self.ordering = ordering
        self.columns = columns

    def is_ordered(self, field: str, way: Literal["asc", "desc"] = "asc") -> bool:
        field_accessor = field.split(".")
        for ordered_field, is_desc in self.ordering:
            if ordered_field == field_accessor:
                return (way == "asc" and is_desc is False) or (
                    way == "desc" and is_desc is True
                )
        return False

    def set_pagination(self, *, limit: int, skip: int) -> "DatatableQueryParameters":
        return DatatableQueryParameters((limit, skip), self.ordering, self.columns)

    def toggle_field_ordering(self, field: str) -> "DatatableQueryParameters":
        field_accessor = field.split(".")
        if self.is_ordered(field, "asc"):
            updated_ordering = [(field_accessor, True)]
        elif self.is_ordered(field, "desc"):
            updated_ordering = []
        else:
            updated_ordering = [(field_accessor, False)]
        return DatatableQueryParameters(
            (self.limit, self.skip), updated_ordering, self.columns
        )

    def toggle_column(self, column: str) -> "DatatableQueryParameters":
        columns = self.columns
        if column in columns:
            columns = [c for c in columns if c != column]
        else:
            columns = columns + [column]
        return DatatableQueryParameters((self.limit, self.skip), self.ordering, columns)

    def __str__(self) -> str:
        ordering_params = [
            f"-{'.'.join(field)}" if is_desc else ".".join(field)
            for (field, is_desc) in self.ordering
        ]
        params = {
            "limit": self.limit,
            "skip": self.skip,
            "ordering": ",".join(ordering_params),
        }
        if self.columns:
            params["columns"] = ",".join(self.columns)
        return urllib.parse.urlencode(params)


class DatatableQueryParametersGetter:
    def __init__(self, default_columns: list[str]) -> None:
        self.default_columns = default_columns

    def __call__(
        self,
        pagination: Pagination = Depends(get_pagination),
        ordering: Ordering = Depends(OrderingGetter()),
        columns: str | None = Query(None),
    ) -> DatatableQueryParameters:
        columns_list = self.default_columns
        if columns is not None:
            columns_list = columns.lower().split(",")
        return DatatableQueryParameters(pagination, ordering, columns_list)



from typing import Any, cast

from fastapi import Depends, HTTPException, Request, status
from fastapi.exceptions import RequestValidationError
from fastapi.security import OAuth2AuthorizationCodeBearer
from fastapi_users import BaseUserManager, InvalidPasswordException
from fastapi_users.authentication import (
    AuthenticationBackend,
    BearerTransport,
    Strategy,
)
from fastapi_users.exceptions import UserAlreadyExists, UserNotExists
from fastapi_users.manager import UUIDIDMixin
from fastapi_users.password import PasswordHelperProtocol
from fastapi_users_db_sqlalchemy import (
    SQLAlchemyBaseOAuthAccountTable,
    SQLAlchemyUserDatabase,
)
from furl import furl
from jwcrypto import jwk
from pydantic import UUID4, ValidationError, create_model
from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select

from fief.crypto.access_token import InvalidAccessToken, read_access_token
from fief.crypto.password import password_helper
from fief.db import AsyncSession
from fief.dependencies.current_workspace import (
    get_current_workspace,
    get_current_workspace_session,
)
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.request import get_request_json
from fief.dependencies.tasks import get_send_task
from fief.dependencies.tenant import (
    get_current_tenant,
    get_tenant_from_create_user_internal,
)
from fief.dependencies.user_field import (
    get_admin_user_create_internal_model,
    get_admin_user_update_model,
    get_user_update_model,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.locale import gettext_lazy as _
from fief.logger import AuditLogger
from fief.models import (
    AuditLogMessage,
    OAuthAccount,
    Tenant,
    User,
    UserField,
    UserFieldValue,
    UserPermission,
    UserRole,
    Workspace,
)
from fief.repositories import (
    OAuthAccountRepository,
    UserPermissionRepository,
    UserRepository,
    UserRoleRepository,
)
from fief.schemas.user import UF, UserCreate, UserCreateInternal, UserRead, UserUpdate
from fief.services.webhooks.models import (
    UserCreated,
    UserForgotPasswordRequested,
    UserPasswordReset,
    UserUpdated,
)
from fief.settings import settings
from fief.tasks import SendTask, on_after_forgot_password, on_after_register


class UserManager(UUIDIDMixin, BaseUserManager[User, UUID4]):
    reset_password_token_secret = settings.secret.get_secret_value()
    verification_token_secret = settings.secret.get_secret_value()

    def __init__(
        self,
        user_db: SQLAlchemyUserDatabase[User, UUID4],
        password_helper: PasswordHelperProtocol,
        workspace: Workspace,
        tenant: Tenant,
        send_task: SendTask,
        audit_logger: AuditLogger,
        trigger_webhooks: TriggerWebhooks,
    ):
        super().__init__(user_db, password_helper)
        self.workspace = workspace
        self.tenant = tenant
        self.send_task = send_task
        self.audit_logger = audit_logger
        self.trigger_webhooks = trigger_webhooks

    async def validate_password(  # type: ignore
        self, password: str, user: UserCreate | User
    ) -> None:
        if len(password) < 8:
            raise InvalidPasswordException(
                reason=_("The password should be at least 8 characters.")
            )

    async def create_with_fields(
        self,
        user_create: UserCreate[UF],
        *,
        user_fields: list[UserField],
        safe: bool = False,
        request: Request | None = None,
    ) -> User:
        await self.validate_password(user_create.password, user_create)

        existing_user = await self.user_db.get_by_email(user_create.email)
        if existing_user is not None:
            raise UserAlreadyExists()

        user_dict = (
            user_create.create_update_dict()
            if safe
            else user_create.create_update_dict_superuser()
        )
        password = user_dict.pop("password")
        user_dict["hashed_password"] = self.password_helper.hash(password)

        user = await self.user_db.create(user_dict)
        await self.user_db.session.refresh(user)  # type: ignore

        for user_field in user_fields:
            user_field_value = UserFieldValue(user_field=user_field)
            try:
                value = user_create.fields.get_value(user_field.slug)
                if value is not None:
                    user_field_value.value = value
                    user.user_field_values.append(user_field_value)
            except AttributeError:
                default = user_field.get_default()
                if default is not None:
                    user_field_value.value = default
                    user.user_field_values.append(user_field_value)

        user = await self.user_db.update(user, {})

        await self.on_after_register(user, request)

        return user

    async def update_with_fields(
        self,
        user_update: UserUpdate[UF],
        user: User,
        *,
        user_fields: list[UserField],
        safe: bool = False,
        request: Request | None = None,
    ) -> User:
        if safe:
            updated_user_data = user_update.create_update_dict()
        else:
            updated_user_data = user_update.create_update_dict_superuser()
        user = await self._update(user, updated_user_data)

        if user_update.fields is not None:
            for user_field in user_fields:
                existing_user_field_value = user.get_user_field_value(user_field)
                # Update existing value
                if existing_user_field_value is not None:
                    try:
                        value = user_update.fields.get_value(user_field.slug)
                        if value is not None:
                            existing_user_field_value.value = value
                    except AttributeError:
                        pass
                # Create new value
                else:
                    user_field_value = UserFieldValue(user_field=user_field)
                    try:
                        value = user_update.fields.get_value(user_field.slug)
                        if value is not None:
                            user_field_value.value = value
                            user.user_field_values.append(user_field_value)
                    except AttributeError:
                        default = user_field.get_default()
                        if default is not None:
                            user_field_value.value = default
                            user.user_field_values.append(user_field_value)

            user = await self.user_db.update(user, {})

        await self.on_after_update(user, updated_user_data, request)

        return user

    async def on_after_register(self, user: User, request: Request | None = None):
        await self.user_db.session.refresh(user)  # type: ignore
        self.audit_logger(AuditLogMessage.USER_REGISTERED, subject_user_id=user.id)
        self.trigger_webhooks(UserCreated, user, UserRead)
        self.send_task(on_after_register, str(user.id), str(self.workspace.id))

    async def on_after_update(
        self, user: User, update_dict: dict[str, Any], request: Request | None = None
    ):
        self.audit_logger(AuditLogMessage.USER_UPDATED, subject_user_id=user.id)
        self.trigger_webhooks(UserUpdated, user, UserRead)

    async def on_after_forgot_password(
        self, user: User, token: str, request: Request | None = None
    ):
        self.audit_logger(
            AuditLogMessage.USER_FORGOT_PASSWORD_REQUESTED, subject_user_id=user.id
        )
        self.trigger_webhooks(UserForgotPasswordRequested, user, UserRead)

        reset_url = furl(self.tenant.url_for(cast(Request, request), "reset:reset"))
        reset_url.add(query_params={"token": token})
        self.send_task(
            on_after_forgot_password,
            str(user.id),
            str(self.workspace.id),
            reset_url.url,
        )

    async def on_after_reset_password(
        self, user: User, request: Request | None = None
    ) -> None:
        self.audit_logger(AuditLogMessage.USER_PASSWORD_RESET, subject_user_id=user.id)
        self.trigger_webhooks(UserPasswordReset, user, UserRead)

    async def on_after_request_verify(
        self, user: User, token: str, request: Request | None = None
    ):
        print(f"Verification requested for user {user.id}. Verification token: {token}")


class SQLAlchemyUserTenantDatabase(SQLAlchemyUserDatabase[User, UUID4]):
    def __init__(
        self,
        session: AsyncSession,
        tenant: Tenant,
        user_table: type[User],
        oauth_account_table: type[SQLAlchemyBaseOAuthAccountTable] | None = None,
    ):
        super().__init__(session, user_table, oauth_account_table=oauth_account_table)
        self.tenant = tenant

    async def _get_user(self, statement: Select) -> User | None:
        statement = statement.where(User.tenant_id == self.tenant.id)
        return await super()._get_user(statement)


class JWTAccessTokenStrategy(Strategy[User, UUID4]):
    def __init__(self, key: jwk.JWK):
        self.key = key

    async def read_token(
        self, token: str | None, user_manager: BaseUserManager[User, UUID4]
    ) -> User | None:
        if token is None:
            return None

        try:
            user_id = read_access_token(self.key, token)
            user = await user_manager.get(user_id)
        except InvalidAccessToken:
            return None
        except UserNotExists:
            return None
        else:
            return user

    async def write_token(self, user: User) -> str:
        raise NotImplementedError()

    async def destroy_token(self, token: str, user: User) -> None:
        raise NotImplementedError()


async def get_jwt_access_token_strategy(
    tenant: Tenant = Depends(get_current_tenant),
) -> JWTAccessTokenStrategy:
    return JWTAccessTokenStrategy(tenant.get_sign_jwk())


async def get_user_db(
    session: AsyncSession = Depends(get_current_workspace_session),
    tenant: Tenant = Depends(get_current_tenant),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, tenant, User)


async def get_user_manager(
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(get_user_db),
    tenant: Tenant = Depends(get_current_tenant),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    return UserManager(
        user_db,
        password_helper,
        workspace,
        tenant,
        send_task,
        audit_logger,
        trigger_webhooks,
    )


async def get_user_db_from_create_user_internal(
    session: AsyncSession = Depends(get_current_workspace_session),
    tenant: Tenant = Depends(get_tenant_from_create_user_internal),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, tenant, User)


async def get_user_manager_from_create_user_internal(
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(
        get_user_db_from_create_user_internal
    ),
    tenant: Tenant = Depends(get_tenant_from_create_user_internal),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    return UserManager(
        user_db,
        password_helper,
        workspace,
        tenant,
        send_task,
        audit_logger,
        trigger_webhooks,
    )


class AuthorizationCodeBearerTransport(BearerTransport):
    scheme: OAuth2AuthorizationCodeBearer  # type: ignore

    def __init__(
        self,
        authorizationUrl: str,
        tokenUrl: str,
        refreshUrl: str,
        scopes: dict[str, str],
    ):
        self.scheme = OAuth2AuthorizationCodeBearer(
            authorizationUrl, tokenUrl, refreshUrl, scopes=scopes
        )


authentication_backend = AuthenticationBackend[User, UUID4](
    "jwt_access_token",
    AuthorizationCodeBearerTransport(
        "/authorize", "/api/token", "/api/token", {"openid": "openid"}
    ),
    get_jwt_access_token_strategy,
)


async def get_paginated_users(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
    get_paginated_objects: GetPaginatedObjects[User] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[User]:
    statement = select(User).options(joinedload(User.tenant))
    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_user_by_id_or_404(
    id: UUID4,
    repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
) -> User:
    user = await repository.get_by_id(id, (joinedload(User.tenant),))

    if user is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return user


async def get_paginated_user_permissions(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    user: User = Depends(get_user_by_id_or_404),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[UserPermission] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[UserPermission]:
    statement = user_permission_repository.get_by_user_statement(user.id)
    return await get_paginated_objects(
        statement, pagination, ordering, user_permission_repository
    )


async def get_user_permissions(
    user: User = Depends(get_user_by_id_or_404),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
) -> list[UserPermission]:
    statement = user_permission_repository.get_by_user_statement(user.id)
    return await user_permission_repository.list(statement)


async def get_paginated_user_roles(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    user: User = Depends(get_user_by_id_or_404),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[UserRole] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[UserRole]:
    statement = user_role_repository.get_by_user_statement(user.id)
    return await get_paginated_objects(
        statement, pagination, ordering, user_role_repository
    )


async def get_user_roles(
    user: User = Depends(get_user_by_id_or_404),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
) -> list[UserRole]:
    statement = user_role_repository.get_by_user_statement(user.id)
    return await user_role_repository.list(statement)


async def get_paginated_user_oauth_accounts(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    user: User = Depends(get_user_by_id_or_404),
    oauth_account_repository: OAuthAccountRepository = Depends(
        get_workspace_repository(OAuthAccountRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[OAuthAccount] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[OAuthAccount]:
    statement = oauth_account_repository.get_by_user_statement(user.id)
    return await get_paginated_objects(
        statement, pagination, ordering, oauth_account_repository
    )


async def get_user_oauth_accounts(
    user: User = Depends(get_user_by_id_or_404),
    oauth_account_repository: OAuthAccountRepository = Depends(
        get_workspace_repository(OAuthAccountRepository)
    ),
) -> list[OAuthAccount]:
    statement = oauth_account_repository.get_by_user_statement(user.id)
    return await oauth_account_repository.list(statement)


async def get_user_db_from_user(
    user: User = Depends(get_user_by_id_or_404),
    session: AsyncSession = Depends(get_current_workspace_session),
) -> SQLAlchemyUserDatabase[User, UUID4]:
    return SQLAlchemyUserTenantDatabase(session, user.tenant, User)


async def get_user_manager_from_user(
    user: User = Depends(get_user_by_id_or_404),
    user_db: SQLAlchemyUserDatabase[User, UUID4] = Depends(get_user_db_from_user),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    return UserManager(
        user_db,
        password_helper,
        workspace,
        user.tenant,
        send_task,
        audit_logger,
        trigger_webhooks,
    )


async def get_user_create_internal(
    json: dict[str, Any] = Depends(get_request_json),
    user_create_internal_model: type[UserCreateInternal[UF]] = Depends(
        get_admin_user_create_internal_model
    ),
) -> UserCreateInternal[UF]:
    body_model = create_model(
        "UserCreateInternalBody",
        body=(user_create_internal_model, ...),
    )
    try:
        validated_user_create_internal = body_model(body=json)
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_create_internal.body  # type: ignore


async def get_user_update(
    json: dict[str, Any] = Depends(get_request_json),
    user_update_model: type[UserUpdate[UF]] = Depends(get_user_update_model),
) -> UserUpdate[UF]:
    body_model = create_model(
        "UserUpdateBody",
        body=(user_update_model, ...),
    )
    try:
        validated_user_update = body_model(body=json)
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_update.body  # type: ignore


async def get_admin_user_update(
    json: dict[str, Any] = Depends(get_request_json),
    user_update_model: type[UserUpdate[UF]] = Depends(get_admin_user_update_model),
) -> UserUpdate[UF]:
    body_model = create_model(
        "UserUpdateBody",
        body=(user_update_model, ...),
    )
    try:
        validated_user_update = body_model(body=json)
    except ValidationError as e:
        raise RequestValidationError(e.raw_errors) from e
    else:
        return validated_user_update.body  # type: ignore



from collections.abc import Mapping
from typing import Any

import asgi_babel
import wtforms.i18n
from asgi_tools import Request
from babel import Locale, support

from fief.paths import LOCALE_DIRECTORY
from fief.settings import settings

Translations = support.Translations


async def select_locale(request: Request) -> str | None:
    user_locale_cookie = request.cookies.get(settings.user_locale_cookie_name)
    if user_locale_cookie is not None:
        return user_locale_cookie
    return await asgi_babel.select_locale_by_request(request)


class BabelMiddleware(asgi_babel.BabelMiddleware):
    def __post_init__(self):
        self.locale_selector = select_locale
        return super().__post_init__()


def get_babel_middleware_kwargs() -> Mapping[str, Any]:
    return dict(locales_dirs=[LOCALE_DIRECTORY, wtforms.i18n.messages_path()])


def get_translations(
    domain: str | None = None, locale: Locale | None = None
) -> support.NullTranslations:
    """
    Load and cache translations.

    Patched version to prevent bug when then first loaded translation is Null.
    """
    from asgi_babel import BABEL, current_locale

    if BABEL is None:
        return support.NullTranslations()

    locale = locale or current_locale.get()
    if not locale:
        return support.NullTranslations()

    domain = domain or BABEL.domain
    locale_code = str(locale)
    if (domain, locale_code) not in BABEL.translations:
        translations = None
        for path in reversed(BABEL.locales_dirs):
            trans = support.Translations.load(path, locales=locale_code, domain=domain)
            if translations and not type(translations) == support.NullTranslations:
                translations._catalog.update(trans._catalog)
            else:
                translations = trans

        BABEL.translations[(domain, locale_code)] = translations

    return BABEL.translations[(domain, locale_code)]


def gettext(string: str, domain: str | None = None, **variables):
    t = get_translations(domain)
    return t.ugettext(string) % variables


def gettext_lazy(string: str, domain: str | None = None, **variables):
    return support.LazyProxy(gettext, string, domain, **variables, enable_cache=False)



from fief.db.engine import create_async_session_maker, create_engine
from fief.settings import settings


def create_main_async_session_maker():
    main_engine = create_engine(settings.get_database_connection_parameters())
    return create_async_session_maker(main_engine)


__all__ = [
    "create_main_async_session_maker",
]



from dataclasses import dataclass

from fief.db.main import create_main_async_session_maker
from fief.db.workspace import WorkspaceEngineManager
from fief.models import Workspace
from fief.repositories import WorkspaceRepository


@dataclass
class WorkspaceStats:
    workspace: Workspace
    reachable: bool
    external_db: bool
    nb_users: int | None = None


class Workspaces:
    def __init__(
        self,
        workspace_repository: WorkspaceRepository,
        workspace_engine_manager: WorkspaceEngineManager,
    ) -> None:
        self.workspace_repository = workspace_repository
        self.workspace_engine_manager = workspace_engine_manager

    async def get_stats(self) -> list[WorkspaceStats]:
        stats: list[WorkspaceStats] = []

        workspaces = await self.workspace_repository.all()
        for workspace in workspaces:
            stats.append(
                WorkspaceStats(
                    workspace=workspace,
                    reachable=True,
                    external_db=workspace.database_type is not None,
                    nb_users=workspace.users_count,
                )
            )

        return stats


async def get_workspaces_stats() -> list[WorkspaceStats]:
    main_async_session_maker = create_main_async_session_maker()
    workspace_engine_manager = WorkspaceEngineManager()
    async with main_async_session_maker() as session:
        workspace_repository = WorkspaceRepository(session)
        workspaces = Workspaces(workspace_repository, workspace_engine_manager)
        return await workspaces.get_stats()



from wtforms import URLField, validators

from fief.forms import CSRFBaseForm, SelectMultipleFieldCheckbox
from fief.services.webhooks.models import WEBHOOK_EVENTS


class BaseWebhookForm(CSRFBaseForm):
    url = URLField(
        "URL",
        validators=[validators.InputRequired(), validators.URL(require_tld=False)],
    )
    events = SelectMultipleFieldCheckbox(
        "Events to notify", choices=[event.key() for event in WEBHOOK_EVENTS]
    )


class WebhookCreateForm(BaseWebhookForm):
    pass


class WebhookUpdateForm(BaseWebhookForm):
    pass



import secrets
from enum import StrEnum

from pydantic import UUID4
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.sqltypes import String

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, ExpiresAt, UUIDModel
from fief.models.oauth_account import OAuthAccount
from fief.models.tenant import Tenant
from fief.settings import settings


class RegistrationSessionFlow(StrEnum):
    PASSWORD = "PASSWORD"
    OAUTH = "OAUTH"


class RegistrationSession(UUIDModel, CreatedUpdatedAt, ExpiresAt, WorkspaceBase):
    __tablename__ = "registration_sessions"
    __lifetime_seconds__ = settings.registration_session_lifetime_seconds

    token: Mapped[str] = mapped_column(
        String(length=255),
        default=secrets.token_urlsafe,
        nullable=False,
        index=True,
        unique=True,
    )

    flow: Mapped[RegistrationSessionFlow] = mapped_column(
        String(length=255), default=RegistrationSessionFlow.PASSWORD, nullable=False
    )

    email: Mapped[str | None] = mapped_column(String(length=320), nullable=True)

    oauth_account_id: Mapped[UUID4 | None] = mapped_column(
        GUID, ForeignKey(OAuthAccount.id, ondelete="CASCADE"), nullable=True
    )
    oauth_account: Mapped[OAuthAccount | None] = relationship(
        "OAuthAccount", lazy="joined"
    )

    tenant_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False
    )
    tenant: Mapped[Tenant] = relationship("Tenant")



from sqlalchemy import select

from fief.models import Role
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class RoleRepository(BaseRepository[Role], UUIDRepositoryMixin[Role]):
    model = Role

    async def get_granted_by_default(self) -> list[Role]:
        statement = select(Role).where(Role.granted_by_default == True)
        return await self.list(statement)



import secrets

from fastapi import APIRouter, Depends, HTTPException, Response, status

from fief import schemas
from fief.crypto.jwk import generate_jwk
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.client import get_client_by_id_or_404, get_paginated_clients
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Client
from fief.repositories import ClientRepository, TenantRepository
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import ClientCreated, ClientDeleted, ClientUpdated

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/", name="clients:list", response_model=PaginatedResults[schemas.client.Client]
)
async def list_clients(
    paginated_clients: PaginatedObjects[Client] = Depends(get_paginated_clients),
) -> PaginatedResults[schemas.client.Client]:
    clients, count = paginated_clients
    return PaginatedResults(
        count=count,
        results=[schemas.client.Client.from_orm(client) for client in clients],
    )


@router.get("/{id:uuid}", name="clients:get", response_model=schemas.client.Client)
async def get_webhook(client: Client = Depends(get_client_by_id_or_404)) -> Client:
    return client


@router.post(
    "/",
    name="clients:create",
    response_model=schemas.client.Client,
    status_code=status.HTTP_201_CREATED,
)
async def create_client(
    client_create: schemas.client.ClientCreate,
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    tenant_repository: TenantRepository = Depends(
        get_workspace_repository(TenantRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.client.Client:
    tenant = await tenant_repository.get_by_id(client_create.tenant_id)
    if tenant is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.CLIENT_CREATE_UNKNOWN_TENANT,
        )

    client = Client(**client_create.dict())
    client = await repository.create(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, client)
    trigger_webhooks(ClientCreated, client, schemas.client.Client)

    return schemas.client.Client.from_orm(client)


@router.patch("/{id:uuid}", name="clients:update", response_model=schemas.client.Client)
async def update_client(
    client_update: schemas.client.ClientUpdate,
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.client.Client:
    client_update_dict = client_update.dict(exclude_unset=True)
    for field, value in client_update_dict.items():
        setattr(client, field, value)

    await repository.update(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, client)
    trigger_webhooks(ClientUpdated, client, schemas.client.Client)

    return schemas.client.Client.from_orm(client)


@router.post(
    "/{id:uuid}/encryption-key",
    name="clients:encryption_key",
    status_code=status.HTTP_201_CREATED,
)
async def create_encryption_key(
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    key = generate_jwk(secrets.token_urlsafe(), "enc")
    client.encrypt_jwk = key.export_public()
    await repository.update(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, client)
    trigger_webhooks(ClientUpdated, client, schemas.client.Client)

    return key.export(as_dict=True)


@router.delete(
    "/{id:uuid}",
    name="clients:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_client(
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, client)
    trigger_webhooks(ClientDeleted, client, schemas.client.Client)



from fastapi import APIRouter, FastAPI
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.staticfiles import StaticFiles

from fief import __version__
from fief.apps.auth.exception_handlers import exception_handlers
from fief.apps.auth.routers.auth import router as auth_router
from fief.apps.auth.routers.oauth import router as oauth_router
from fief.apps.auth.routers.register import router as register_router
from fief.apps.auth.routers.reset import router as reset_router
from fief.apps.auth.routers.token import router as token_router
from fief.apps.auth.routers.user import router as user_router
from fief.apps.auth.routers.well_known import router as well_known_router
from fief.locale import BabelMiddleware, get_babel_middleware_kwargs
from fief.middlewares.cors import CORSMiddlewarePath
from fief.middlewares.csrf import CSRFCookieSetterMiddleware
from fief.middlewares.security_headers import SecurityHeadersMiddleware
from fief.paths import STATIC_DIRECTORY


def include_routers(router: APIRouter) -> APIRouter:
    router.include_router(auth_router, include_in_schema=False)
    router.include_router(register_router, include_in_schema=False)
    router.include_router(reset_router, include_in_schema=False)
    router.include_router(token_router, prefix="/api")
    router.include_router(user_router, prefix="/api")
    router.include_router(well_known_router, prefix="/.well-known")

    return router


default_tenant_router = include_routers(APIRouter())
tenant_router = include_routers(APIRouter(prefix="/{tenant_slug}"))


app = FastAPI(title="Fief Authentication API", version=__version__)
app.add_middleware(SecurityHeadersMiddleware)
app.add_middleware(CSRFCookieSetterMiddleware)
app.add_middleware(GZipMiddleware)
app.add_middleware(
    CORSMiddlewarePath,
    path_regex=r"^/api|^/\.well-known",
    allow_origins=["*"],
    allow_credentials=False,
    allow_methods=["*"],
    allow_headers=["Authorization", "X-Requested-With"],
)

app.add_middleware(BabelMiddleware, **get_babel_middleware_kwargs())
app.include_router(oauth_router, include_in_schema=False)
app.include_router(default_tenant_router)
app.include_router(tenant_router)
app.mount("/static", StaticFiles(directory=STATIC_DIRECTORY), name="auth:static")

for exc, handler in exception_handlers.items():
    app.add_exception_handler(exc, handler)


__all__ = ["app"]



from collections.abc import Mapping
from datetime import date, datetime
from enum import Enum
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, StrictBool, StrictInt, StrictStr, constr
from pydantic.generics import GenericModel

from fief.models import UserField as UserFieldModel
from fief.models import UserFieldType
from fief.schemas.generics import (
    Address,
    CreatedUpdatedAt,
    PhoneNumber,
    Timezone,
    UUIDSchema,
)

USER_FIELD_TYPE_MAP: Mapping[UserFieldType, type[Any]] = {
    UserFieldType.STRING: constr(min_length=1),
    UserFieldType.INTEGER: int,
    UserFieldType.BOOLEAN: bool,
    UserFieldType.DATE: date,
    UserFieldType.DATETIME: datetime,
    UserFieldType.CHOICE: str,
    UserFieldType.PHONE_NUMBER: PhoneNumber,
    UserFieldType.ADDRESS: Address,
    UserFieldType.TIMEZONE: Timezone,
}

USER_FIELD_CAN_HAVE_DEFAULT: Mapping[UserFieldType, bool] = {
    UserFieldType.STRING: True,
    UserFieldType.INTEGER: True,
    UserFieldType.BOOLEAN: True,
    UserFieldType.DATE: False,
    UserFieldType.DATETIME: False,
    UserFieldType.CHOICE: True,
    UserFieldType.PHONE_NUMBER: False,
    UserFieldType.ADDRESS: False,
    UserFieldType.TIMEZONE: True,
}

D = TypeVar("D", Timezone, bool, int, str)
UFT = TypeVar("UFT", bound=UserFieldType)


def get_user_field_pydantic_type(field: UserFieldModel) -> type[Any]:
    if field.type == UserFieldType.CHOICE:
        choices = (
            field.configuration["choices"] if field.configuration["choices"] else []
        )
        return Enum(  # type: ignore
            f"{field.slug.capitalize()}Enum",
            [(value, value) for (value, _) in choices],
            type=str,
        )
    return USER_FIELD_TYPE_MAP[field.type]


class UserFieldConfigurationBase(BaseModel):
    at_registration: bool
    required: bool
    at_update: bool


class ChoiceTupleType(tuple[str, str]):
    pass


class UserFieldConfiguration(UserFieldConfigurationBase):
    choices: list[ChoiceTupleType] | None
    default: Timezone | StrictBool | StrictInt | StrictStr | None


class UserFieldConfigurationDefault(
    GenericModel, Generic[D], UserFieldConfigurationBase
):
    default: D | None


class UserFieldConfigurationChoice(UserFieldConfigurationDefault[str]):
    choices: list[ChoiceTupleType] | None


class UserFieldCreate(BaseModel):
    name: str
    slug: str
    type: UserFieldType
    configuration: UserFieldConfiguration


class UserFieldUpdate(BaseModel):
    name: str | None
    slug: str | None
    configuration: UserFieldConfiguration | None


class BaseUserField(UUIDSchema, CreatedUpdatedAt):
    name: str
    slug: str
    type: UserFieldType
    configuration: UserFieldConfiguration


class UserField(BaseUserField):
    pass



from fastapi import APIRouter, Depends, HTTPException, Response, status

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.permission import (
    get_paginated_permissions,
    get_permission_by_id_or_404,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Permission
from fief.repositories import PermissionRepository
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import (
    PermissionCreated,
    PermissionDeleted,
    PermissionUpdated,
)

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/",
    name="permissions:list",
    response_model=PaginatedResults[schemas.permission.Permission],
)
async def list_permissions(
    paginated_permissions: PaginatedObjects[Permission] = Depends(
        get_paginated_permissions
    ),
) -> PaginatedResults[schemas.permission.Permission]:
    permissions, count = paginated_permissions
    return PaginatedResults(
        count=count,
        results=[
            schemas.permission.Permission.from_orm(permission)
            for permission in permissions
        ],
    )


@router.get(
    "/{id:uuid}", name="permissions:get", response_model=schemas.permission.Permission
)
async def get_permission(
    permission: Permission = Depends(get_permission_by_id_or_404),
) -> Permission:
    return permission


@router.post(
    "/",
    name="permissions:create",
    response_model=schemas.permission.Permission,
    status_code=status.HTTP_201_CREATED,
)
async def create_permission(
    permission_create: schemas.permission.PermissionCreate,
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.permission.Permission:
    existing_permission = await repository.get_by_codename(permission_create.codename)
    if existing_permission is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.PERMISSION_CREATE_CODENAME_ALREADY_EXISTS,
        )

    permission = Permission(**permission_create.dict())
    permission = await repository.create(permission)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, permission)
    trigger_webhooks(PermissionCreated, permission, schemas.permission.Permission)

    return schemas.permission.Permission.from_orm(permission)


@router.patch(
    "/{id:uuid}",
    name="permissions:update",
    response_model=schemas.permission.Permission,
)
async def update_permission(
    permission_update: schemas.permission.PermissionUpdate,
    permission: Permission = Depends(get_permission_by_id_or_404),
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.permission.Permission:
    updated_codename = permission_update.codename
    if updated_codename is not None and updated_codename != permission.codename:
        existing_permission = await repository.get_by_codename(updated_codename)
        if existing_permission is not None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.PERMISSION_UPDATE_CODENAME_ALREADY_EXISTS,
            )

    permission_update_dict = permission_update.dict(exclude_unset=True)
    for field, value in permission_update_dict.items():
        setattr(permission, field, value)

    await repository.update(permission)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, permission)
    trigger_webhooks(PermissionUpdated, permission, schemas.permission.Permission)

    return schemas.permission.Permission.from_orm(permission)


@router.delete(
    "/{id:uuid}",
    name="permissions:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_permission(
    permission: Permission = Depends(get_permission_by_id_or_404),
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(permission)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, permission)
    trigger_webhooks(PermissionDeleted, permission, schemas.permission.Permission)



from datetime import datetime

from pydantic import UUID4

from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema
from fief.schemas.oauth_provider import OAuthProvider


class OAuthAccount(UUIDSchema, CreatedUpdatedAt):
    account_id: str
    oauth_provider_id: UUID4
    oauth_provider: OAuthProvider


class OAuthAccountAccessToken(UUIDSchema):
    account_id: str
    access_token: str
    expires_at: datetime | None



from pydantic import UUID4
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.role import Role
from fief.models.user import User


class UserRole(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_roles"
    __table_args__ = (UniqueConstraint("user_id", "role_id"),)

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    role_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Role.id, ondelete="CASCADE"), nullable=False
    )

    user: Mapped[User] = relationship("User")
    role: Mapped[Role] = relationship("Role")

    def __repr__(self) -> str:
        return f"UserRole(id={self.id}, user_id={self.user_id}, role_id={self.role_id})"



from typing import Any

from pydantic import HttpUrl, SecretStr, root_validator

from fief.errors import APIErrorCode
from fief.schemas.generics import BaseModel, CreatedUpdatedAt, UUIDSchema
from fief.services.oauth_provider import AvailableOAuthProvider


def validate_openid_provider(cls, values: dict[str, Any]):
    provider: AvailableOAuthProvider = values["provider"]
    if provider == AvailableOAuthProvider.OPENID:
        openid_configuration_endpoint = values.get("openid_configuration_endpoint")
        if openid_configuration_endpoint is None:
            raise ValueError(
                APIErrorCode.OAUTH_PROVIDER_MISSING_OPENID_CONFIGURATION_ENDPOINT.value
            )
    return values


class OAuthProviderCreate(BaseModel):
    provider: AvailableOAuthProvider
    client_id: str
    client_secret: str
    scopes: list[str]
    name: str | None = None
    openid_configuration_endpoint: HttpUrl | None = None

    _validate_openid_provider = root_validator(allow_reuse=True)(
        validate_openid_provider
    )


class OAuthProviderUpdate(BaseModel):
    client_id: str | None
    client_secret: str | None
    scopes: list[str] | None
    name: str | None
    openid_configuration_endpoint: HttpUrl | None


class OAuthProviderUpdateProvider(OAuthProviderUpdate):
    provider: AvailableOAuthProvider

    _validate_openid_provider = root_validator(allow_reuse=True)(
        validate_openid_provider
    )


class BaseOAuthProvider(UUIDSchema, CreatedUpdatedAt):
    provider: AvailableOAuthProvider
    client_id: SecretStr
    client_secret: SecretStr
    scopes: list[str]
    name: str | None = None
    openid_configuration_endpoint: HttpUrl | None = None


class OAuthProvider(BaseOAuthProvider):
    pass


class OAuthProviderEmbedded(BaseOAuthProvider):
    pass



from datetime import datetime, timezone

from pydantic import UUID4
from sqlalchemy import ForeignKey, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.crypto.encryption import FernetEngine, StringEncryptedType
from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.oauth_provider import OAuthProvider
from fief.models.tenant import Tenant
from fief.models.user import User
from fief.settings import settings


class OAuthAccount(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "oauth_accounts"
    __table_args__ = (UniqueConstraint("oauth_provider_id", "user_id"),)

    access_token: Mapped[str] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=False
    )
    expires_at: Mapped[datetime | None] = mapped_column(
        TIMESTAMPAware(timezone=True), nullable=True
    )
    refresh_token: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    account_id: Mapped[str] = mapped_column(
        String(length=512), index=True, nullable=False
    )
    account_email: Mapped[str | None] = mapped_column(String(length=512), nullable=True)

    oauth_provider_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(OAuthProvider.id, ondelete="CASCADE"), nullable=False
    )
    oauth_provider: Mapped[OAuthProvider] = relationship("OAuthProvider", lazy="joined")

    user_id: Mapped[UUID4 | None] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=True
    )
    user: Mapped[User | None] = relationship("User", lazy="joined")

    tenant_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False
    )
    tenant: Mapped[Tenant] = relationship("Tenant")

    def is_expired(self) -> bool:
        return self.expires_at is not None and self.expires_at < datetime.now(
            timezone.utc
        )



from typing import Any

from pydantic import BaseModel
from pydantic.fields import Field


class OAuthError(BaseModel):
    error: str = Field(
        ...,
        regex="invalid_tenant|invalid_provider|oauth_error|invalid_session|missing_code|access_token_error|id_email_error|inactive_user|user_already_exists",
    )
    error_description: Any | None = None
    error_uri: str | None = None

    @classmethod
    def get_invalid_tenant(cls, error_description: Any | None = None):
        return cls(error="invalid_tenant", error_description=error_description)

    @classmethod
    def get_invalid_provider(cls, error_description: Any | None = None):
        return cls(error="invalid_provider", error_description=error_description)

    @classmethod
    def get_oauth_error(cls, error_description: Any | None = None):
        return cls(error="oauth_error", error_description=error_description)

    @classmethod
    def get_invalid_session(cls, error_description: Any | None = None):
        return cls(error="invalid_session", error_description=error_description)

    @classmethod
    def get_missing_code(cls, error_description: Any | None = None):
        return cls(error="missing_code", error_description=error_description)

    @classmethod
    def get_access_token_error(cls, error_description: Any | None = None):
        return cls(error="access_token_error", error_description=error_description)

    @classmethod
    def get_id_email_error(cls, error_description: Any | None = None):
        return cls(error="id_email_error", error_description=error_description)

    @classmethod
    def get_inactive_user(cls, error_description: Any | None = None):
        return cls(error="inactive_user", error_description=error_description)

    @classmethod
    def get_user_already_exists(cls, error_description: Any | None = None):
        return cls(error="user_already_exists", error_description=error_description)



from fastapi import Depends

from fief.dependencies.permission import (
    UserPermissionsGetter,
    get_user_permissions_getter,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.repositories import (
    AuthorizationCodeRepository,
    GrantRepository,
    LoginSessionRepository,
    SessionTokenRepository,
)
from fief.services.authentication_flow import AuthenticationFlow


async def get_authentication_flow(
    authorization_code_repository: AuthorizationCodeRepository = Depends(
        get_workspace_repository(AuthorizationCodeRepository)
    ),
    login_session_repository: LoginSessionRepository = Depends(
        get_workspace_repository(LoginSessionRepository)
    ),
    session_token_repository: SessionTokenRepository = Depends(
        get_workspace_repository(SessionTokenRepository)
    ),
    grant_repository: GrantRepository = Depends(
        get_workspace_repository(GrantRepository)
    ),
    get_user_permissions: UserPermissionsGetter = Depends(get_user_permissions_getter),
) -> AuthenticationFlow:
    return AuthenticationFlow(
        authorization_code_repository,
        login_session_repository,
        session_token_repository,
        grant_repository,
        get_user_permissions,
    )



from pydantic import UUID4
from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, ExpiresAt, UUIDModel
from fief.models.user import User
from fief.settings import settings


class SessionToken(UUIDModel, CreatedUpdatedAt, ExpiresAt, WorkspaceBase):
    __tablename__ = "session_tokens"
    __lifetime_seconds__ = settings.session_lifetime_seconds

    token: Mapped[str] = mapped_column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    user: Mapped[User] = relationship("User", lazy="joined")



import base64
import hashlib
import secrets


def get_code_verifier_hash(verifier: str) -> str:
    hasher = hashlib.sha256()
    hasher.update(verifier.encode("ascii"))
    digest = hasher.digest()
    b64_digest = base64.urlsafe_b64encode(digest).decode("ascii")
    return b64_digest[:-1]  # Remove the padding "=" at the end


def verify_code_verifier(verifier: str, challenge: str, method: str) -> bool:
    if method == "plain":
        return secrets.compare_digest(verifier, challenge)
    elif method == "S256":
        verifier_hash = get_code_verifier_hash(verifier)
        return secrets.compare_digest(verifier_hash, challenge)
    return False  # pragma: no cover



"""Add EmailTemplate model

Revision ID: 2501f74d0c55
Revises: 67c0904f3ebc
Create Date: 2022-10-28 16:00:00.652479

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "2501f74d0c55"
down_revision = "67c0904f3ebc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_email_templates",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "created_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("type", sa.String(length=255), nullable=False),
        sa.Column("subject", sa.Text(), nullable=False),
        sa.Column("content", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("type"),
    )
    op.create_index(
        op.f("ix_fief_email_templates_created_at"),
        "fief_email_templates",
        ["created_at"],
        unique=False,
    )
    op.create_index(
        op.f("ix_fief_email_templates_updated_at"),
        "fief_email_templates",
        ["updated_at"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_fief_email_templates_updated_at"), table_name="fief_email_templates"
    )
    op.drop_index(
        op.f("ix_fief_email_templates_created_at"), table_name="fief_email_templates"
    )
    op.drop_table("fief_email_templates")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING

from sqlalchemy import Enum, Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.crypto.encryption import FernetEngine, StringEncryptedType
from fief.db.types import (
    DatabaseConnectionParameters,
    DatabaseType,
    create_database_connection_parameters,
)
from fief.models.base import MainBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel
from fief.settings import settings

if TYPE_CHECKING:
    from fief.models.workspace_user import WorkspaceUser


class Workspace(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "workspaces"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    domain: Mapped[str] = mapped_column(String(length=255), nullable=False)

    database_type: Mapped[DatabaseType | None] = mapped_column(
        Enum(DatabaseType), nullable=True
    )
    database_host: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    database_port: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    database_username: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    database_password: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    database_name: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )
    database_ssl_mode: Mapped[str | None] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=True
    )

    alembic_revision: Mapped[str | None] = mapped_column(
        String(length=255), nullable=True, index=True
    )

    users_count: Mapped[int] = mapped_column(
        Integer, nullable=False, default=0, server_default="0"
    )

    workspace_users: Mapped[list["WorkspaceUser"]] = relationship(
        "WorkspaceUser", back_populates="workspace", cascade="all, delete"
    )

    def __repr__(self) -> str:
        return f"Workspace(id={self.id}, name={self.name}, domain={self.domain})"

    def get_database_connection_parameters(
        self, asyncio=True
    ) -> DatabaseConnectionParameters:
        """
        Return the database URL and connection arguments for this workspace.

        If it's not specified on the model, the instance database URL is returned.
        """
        if self.database_type is None:
            url = settings.get_database_connection_parameters(
                asyncio, schema=self.get_schema_name()
            )
        else:
            url = create_database_connection_parameters(
                self.database_type,
                asyncio=asyncio,
                username=self.database_username,
                password=self.database_password,
                host=self.database_host,
                port=int(self.database_port) if self.database_port else None,
                database=self.database_name,
                path=settings.database_location,
                schema=self.get_schema_name(),
                ssl_mode=self.database_ssl_mode,
            )

        return url

    def get_schema_name(self) -> str:
        """
        Return the SQL schema name where the data is stored.
        """
        return str(self.id)



from pydantic import UUID4
from sqlalchemy import delete, select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import Select

from fief.models import UserPermission
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class UserPermissionRepository(
    BaseRepository[UserPermission], UUIDRepositoryMixin[UserPermission]
):
    model = UserPermission

    def get_by_user_statement(
        self, user: UUID4, *, direct_only: bool = False
    ) -> Select:
        statement = (
            select(UserPermission)
            .where(UserPermission.user_id == user)
            .options(
                joinedload(UserPermission.permission),
                joinedload(UserPermission.from_role),
            )
        )

        if direct_only:
            statement = statement.where(UserPermission.from_role == None)

        return statement

    async def get_by_permission_and_user(
        self, user: UUID4, permission: UUID4, *, direct_only: bool = False
    ) -> UserPermission | None:
        statement = (
            select(UserPermission)
            .where(
                UserPermission.user_id == user,
                UserPermission.permission_id == permission,
            )
            .options(
                joinedload(UserPermission.permission),
                joinedload(UserPermission.from_role),
            )
        )

        if direct_only:
            statement = statement.where(UserPermission.from_role == None)

        return await self.get_one_or_none(statement)

    async def delete_by_user_and_role(self, user: UUID4, from_role: UUID4) -> None:
        statement = delete(UserPermission).where(
            UserPermission.user_id == user, UserPermission.from_role_id == from_role
        )
        await self._execute_statement(statement)

    async def delete_by_permission_and_role(
        self, permission: UUID4, from_role: UUID4
    ) -> None:
        statement = delete(UserPermission).where(
            UserPermission.permission_id == permission,
            UserPermission.from_role_id == from_role,
        )
        await self._execute_statement(statement)

    async def delete_by_role(self, from_role: UUID4) -> None:
        statement = delete(UserPermission).where(
            UserPermission.from_role_id == from_role
        )
        await self._execute_statement(statement)



import random
import string
import uuid

from slugify import slugify
from sqlalchemy import select

from fief.models import Workspace, WorkspaceUser
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin
from fief.settings import settings


class WorkspaceRepository(BaseRepository[Workspace], UUIDRepositoryMixin[Workspace]):
    model = Workspace

    async def get_by_admin_user(self, user_id: uuid.UUID) -> list[Workspace]:
        statement = (
            select(Workspace)
            .join(Workspace.workspace_users)
            .where(WorkspaceUser.user_id == user_id)
        )
        return await self.list(statement)

    async def get_by_domain(self, domain: str) -> Workspace | None:
        statement = select(Workspace).where(Workspace.domain == domain)
        return await self.get_one_or_none(statement)

    async def get_main(self) -> Workspace | None:
        return await self.get_by_domain(settings.fief_domain)

    async def get_available_subdomain(self, name: str) -> str:
        slug = slugify(name)
        domain = f"{slug}.{settings.root_domain}"
        workspace = await self.get_by_domain(domain)

        if workspace is None:
            return domain

        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=6)
        )
        return f"{slug}-{random_string}.{settings.root_domain}"



from enum import StrEnum
from typing import TYPE_CHECKING, Any, TypedDict

from sqlalchemy import JSON, String
from sqlalchemy import Enum as SQLEnum
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel

if TYPE_CHECKING:
    from fief.models.user_field_value import UserFieldValue


class UserFieldNotChoice(TypeError):
    pass


class UserFieldChoiceNotExistingValue(ValueError):
    pass


class UserFieldType(StrEnum):
    STRING = "STRING"
    INTEGER = "INTEGER"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    DATETIME = "DATETIME"
    CHOICE = "CHOICE"
    PHONE_NUMBER = "PHONE_NUMBER"
    ADDRESS = "ADDRESS"
    TIMEZONE = "TIMEZONE"

    def get_display_name(self) -> str:
        display_names = {
            UserFieldType.STRING: "String",
            UserFieldType.INTEGER: "Integer",
            UserFieldType.BOOLEAN: "Boolean",
            UserFieldType.DATE: "Date",
            UserFieldType.DATETIME: "Date & Time",
            UserFieldType.CHOICE: "Choice",
            UserFieldType.PHONE_NUMBER: "Phone number",
            UserFieldType.ADDRESS: "Address",
            UserFieldType.TIMEZONE: "Timezone",
        }
        return display_names[self]

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [(member.value, member.get_display_name()) for member in cls]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


class UserFieldConfiguration(TypedDict):
    choices: list[tuple[str, str]] | None
    at_registration: bool
    at_update: bool
    required: bool
    default: Any | None


class UserField(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_fields"

    name: Mapped[str] = mapped_column(String(length=320), nullable=False)
    slug: Mapped[str] = mapped_column(
        String(length=320), index=True, nullable=False, unique=True
    )
    type: Mapped[UserFieldType] = mapped_column(
        SQLEnum(UserFieldType), index=True, nullable=True
    )
    configuration: Mapped[UserFieldConfiguration] = mapped_column(JSON, nullable=False)

    user_field_values: Mapped[list["UserFieldValue"]] = relationship(
        "UserFieldValue", back_populates="user_field", cascade="all, delete"
    )

    def get_required(self) -> bool:
        return self.configuration["required"]

    def get_default(self) -> Any | None:
        return self.configuration["default"]

    def get_type_display_name(self) -> str:
        return UserFieldType[self.type].get_display_name()

    def get_choice_label(self, value: str) -> str:
        if self.type != UserFieldType.CHOICE:
            raise UserFieldNotChoice()
        for value, label in self.configuration["choices"] or []:
            if value == value:
                return label
        raise UserFieldChoiceNotExistingValue()



import binascii

from cryptography.fernet import Fernet
from sqlalchemy import Text
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    FernetEngine as BaseFernetEngine,
)
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    StringEncryptedType as BaseStringEncryptedType,
)


class StringEncryptedType(BaseStringEncryptedType):
    impl = Text


def generate_key() -> bytes:
    return Fernet.generate_key()


def is_valid_key(key: bytes) -> bool:
    try:
        Fernet(key)
    except binascii.Error:
        return False
    else:
        return True


class FernetEngine(BaseFernetEngine):
    """
    Overload of the built-in SQLAlchemy Utils Fernet engine.

    For unknown reasons, they hash the encryption key before using it.

    For backward compatibility, we adjust the implementation to use the key as provided.
    """

    def _update_key(self, key: bytes):
        return self._initialize_engine(key)

    def _initialize_engine(self, parent_class_key: bytes):
        self.secret_key = parent_class_key
        self.fernet = Fernet(self.secret_key)



from postmarker.core import PostmarkClient
from postmarker.exceptions import ClientError

from fief.services.email.base import EmailProvider, SendEmailError, format_address


class Postmark(EmailProvider):
    def __init__(self, server_token: str) -> None:
        self._client = PostmarkClient(server_token=server_token)

    def send_email(
        self,
        *,
        sender: tuple[str, str | None],
        recipient: tuple[str, str | None],
        subject: str,
        html: str | None = None,
        text: str | None = None,
    ):
        from_email, from_name = sender
        to_email, to_name = recipient
        try:
            self._client.emails.send(
                From=format_address(from_email, from_name),
                To=format_address(to_email, to_name),
                Subject=subject,
                HtmlBody=html,
                TextBody=text,
            )
        except ClientError as e:
            raise SendEmailError(str(e)) from e



from collections.abc import AsyncGenerator

from fief.db import AsyncSession
from fief.db.main import create_main_async_session_maker
from fief.db.workspace import WorkspaceEngineManager

workspace_engine_manager = WorkspaceEngineManager()
main_async_session_maker = create_main_async_session_maker()


async def get_workspace_engine_manager() -> WorkspaceEngineManager:
    return workspace_engine_manager


async def get_main_async_session() -> AsyncGenerator[AsyncSession, None]:
    async with main_async_session_maker() as session:
        yield session



from fief.apps.api.app import app as api_app
from fief.apps.auth.app import app as auth_app
from fief.apps.dashboard.app import app as dashboard_app

__all__ = ["api_app", "auth_app", "dashboard_app"]



from wtforms import EmailField, HiddenField, PasswordField, validators

from fief.forms import CSRFBaseForm
from fief.locale import gettext_lazy as _


class ForgotPasswordForm(CSRFBaseForm):
    email = EmailField(
        _("Email address"), validators=[validators.InputRequired(), validators.Email()]
    )


class ResetPasswordForm(CSRFBaseForm):
    password = PasswordField(_("New password"), validators=[validators.InputRequired()])
    token = HiddenField(validators=[validators.InputRequired()])



import sentry_dramatiq
import sentry_sdk
from sentry_sdk.integrations.redis import RedisIntegration

from fief import __version__
from fief.logger import logger
from fief.settings import settings

sentry_sdk.init(
    dsn=settings.sentry_dsn_worker,
    environment=settings.environment.value,
    release=__version__,
    integrations=[sentry_dramatiq.DramatiqIntegration(), RedisIntegration()],
)

from fief import tasks  # noqa: E402

logger.info("Fief Worker started", version=__version__)

__all__ = ["tasks"]



from wtforms import (
    EmailField,
    FieldList,
    Form,
    FormField,
    PasswordField,
    StringField,
    validators,
    widgets,
)

from fief.forms import (
    ComboboxSelectField,
    CSRFBaseForm,
    empty_string_to_none,
    get_form_field,
)
from fief.models import UserField


class BaseUserForm(CSRFBaseForm):
    email = EmailField(
        "Email address", validators=[validators.InputRequired(), validators.Email()]
    )

    @classmethod
    async def get_form_class(
        cls, user_fields: list[UserField]
    ) -> type["UserCreateForm"]:
        class UserFormFields(Form):
            pass

        for field in user_fields:
            setattr(UserFormFields, field.slug, get_form_field(field))

        class UserForm(cls):  # type: ignore
            fields = FormField(UserFormFields)

        return UserForm


class UserCreateForm(BaseUserForm):
    password = PasswordField(
        "Password",
        validators=[validators.InputRequired()],
        widget=widgets.PasswordInput(hide_value=False),
    )
    tenant = ComboboxSelectField(
        "Tenant",
        query_endpoint_path="/admin/tenants/",
        validators=[validators.InputRequired(), validators.UUID()],
    )


class UserUpdateForm(BaseUserForm):
    password = PasswordField(
        "Password",
        filters=(empty_string_to_none,),
        widget=widgets.PasswordInput(hide_value=False),
    )


class UserAccessTokenForm(CSRFBaseForm):
    client = ComboboxSelectField(
        "Client",
        description="The access token will be tied to this client.",
        query_endpoint_path="/admin/clients/",
        validators=[validators.InputRequired(), validators.UUID()],
    )
    scopes = FieldList(
        StringField(validators=[validators.InputRequired()]),
        label="Scopes",
        default=["openid"],
    )


class CreateUserPermissionForm(CSRFBaseForm):
    permission = ComboboxSelectField(
        "Add new permission",
        query_endpoint_path="/admin/access-control/permissions/",
        validators=[validators.InputRequired(), validators.UUID()],
    )


class CreateUserRoleForm(CSRFBaseForm):
    role = ComboboxSelectField(
        "Add new role",
        query_endpoint_path="/admin/access-control/roles/",
        validators=[validators.InputRequired(), validators.UUID()],
    )



from typing import Any
from urllib.parse import urlsplit, urlunsplit

from fief_client import FiefAsync

from fief.settings import settings


class FiefAsyncRelativeEndpoints(FiefAsync):
    def _get_endpoint_url(
        self, openid_configuration: dict[str, Any], field: str
    ) -> str:
        """
        Tweak the URL so they are relative instead of absolute.

        Fief will always return absolute URL on the .well-known endpoint.

        For the Fief-ception, we always need to call them from localhost,
        so we just get rid of the scheme and host part.

        The only exception is `authorization_endpoint`:
        in this case, we want it to be absolute because it's returned to the
        user and loaded by its browser on the outside.
        """
        netloc = settings.fief_domain if field == "authorization_endpoint" else ""
        url = super()._get_endpoint_url(openid_configuration, field)
        return urlunsplit(("", netloc, *urlsplit(url)[2:]))


fief = FiefAsyncRelativeEndpoints(
    f"http://localhost:{settings.port}",  # Always call Fief on localhost
    settings.fief_client_id,
    settings.fief_client_secret,
    encryption_key=settings.fief_encryption_key,
    host=settings.fief_domain,  # Manually set the host so the workspace is recognized
)


async def get_fief() -> FiefAsyncRelativeEndpoints:
    """
    This is Fief-ception.

    We are configuring a Fief client to authenticate Fief users to their account.
    """
    return fief



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from fief import __version__
from fief.apps.api.routers.clients import router as clients_router
from fief.apps.api.routers.email_templates import router as email_templates_router
from fief.apps.api.routers.oauth_providers import router as oauth_providers_router
from fief.apps.api.routers.permissions import router as permissions_router
from fief.apps.api.routers.roles import router as roles_router
from fief.apps.api.routers.tenants import router as tenants_router
from fief.apps.api.routers.user_fields import router as user_fields_router
from fief.apps.api.routers.users import router as users_router
from fief.apps.api.routers.webhooks import router as webhooks_router
from fief.middlewares.security_headers import SecurityHeadersMiddleware
from fief.services.localhost import is_localhost
from fief.settings import settings

app = FastAPI(
    title="Fief Administration API",
    version=__version__,
    contact={
        "name": "Fief",
        "url": "https://docs.fief.dev",
        "email": "contact@fief.dev",
    },
    license_info={
        "name": "Elastic License 2.0 (ELv2)",
        "url": "https://github.com/fief-dev/fief/blob/main/LICENSE.md",
    },
    servers=[
        {
            "url": "{scheme}://{host}/admin/api",
            "variables": {
                "scheme": {
                    "enum": ["http", "https"],
                    "default": "http"
                    if is_localhost(settings.fief_domain)
                    else "https",
                },
                "host": {
                    "default": settings.fief_domain,
                    "description": "Your Fief's workspace hostname",
                },
            },
        }
    ],
)

app.add_middleware(
    CORSMiddleware,
    allow_origin_regex=settings.allow_origin_regex,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(SecurityHeadersMiddleware)

app.include_router(clients_router, prefix="/clients", tags=["Clients"])
app.include_router(
    email_templates_router, prefix="/email-templates", tags=["Email templates"]
)
app.include_router(
    oauth_providers_router, prefix="/oauth-providers", tags=["OAuth Providers"]
)
app.include_router(permissions_router, prefix="/permissions", tags=["Permissions"])
app.include_router(roles_router, prefix="/roles", tags=["Roles"])
app.include_router(tenants_router, prefix="/tenants", tags=["Tenants"])
app.include_router(users_router, prefix="/users", tags=["Users"])
app.include_router(user_fields_router, prefix="/user-fields", tags=["User fields"])
app.include_router(webhooks_router, prefix="/webhooks", tags=["Webhooks"])

__all__ = ["app"]



from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from fief import schemas
from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.models import Tenant
from fief.repositories import TenantRepository


async def get_current_tenant(
    tenant_slug: str | None = Query(None),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
) -> Tenant:
    if tenant_slug is None:
        tenant = await repository.get_default()
    else:
        tenant = await repository.get_by_slug(tenant_slug)

    if tenant is None:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)

    return tenant


async def get_tenant_from_create_user_internal(
    user_create: schemas.user.UserCreateInternal,
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
) -> Tenant:
    tenant = await repository.get_by_id(user_create.tenant_id)

    if tenant is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_CREATE_UNKNOWN_TENANT,
        )

    return tenant


async def get_paginated_tenants(
    query: str | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
    get_paginated_objects: GetPaginatedObjects[Tenant] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Tenant]:
    statement = select(Tenant)

    if query is not None:
        statement = statement.where(Tenant.name.ilike(f"%{query}%"))

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_tenant_by_id_or_404(
    id: UUID4,
    repository: TenantRepository = Depends(get_workspace_repository(TenantRepository)),
) -> Tenant:
    tenant = await repository.get_by_id(id, (selectinload(Tenant.theme),))

    if tenant is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return tenant



from sqlalchemy import Boolean, Integer, String
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class Theme(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "themes"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    default: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)

    primary_color: Mapped[str] = mapped_column(String(length=255), nullable=False)
    primary_color_hover: Mapped[str] = mapped_column(String(length=255), nullable=False)
    primary_color_light: Mapped[str] = mapped_column(String(length=255), nullable=False)

    input_color: Mapped[str] = mapped_column(String(length=255), nullable=False)
    input_color_background: Mapped[str] = mapped_column(
        String(length=255), nullable=False
    )

    light_color: Mapped[str] = mapped_column(String(length=255), nullable=False)
    light_color_hover: Mapped[str] = mapped_column(String(length=255), nullable=False)

    text_color: Mapped[str] = mapped_column(String(length=255), nullable=False)
    accent_color: Mapped[str] = mapped_column(String(length=255), nullable=False)

    background_color: Mapped[str] = mapped_column(String(length=255), nullable=False)

    font_size: Mapped[int] = mapped_column(Integer, nullable=False)
    font_family: Mapped[str] = mapped_column(String(length=255), nullable=False)
    font_css_url: Mapped[str | None] = mapped_column(
        String(length=512), default=None, nullable=True
    )

    @classmethod
    def build_default(cls) -> "Theme":
        return cls(
            name="Default",
            default=True,
            primary_color="#f43f5e",
            primary_color_hover="#e11d48",
            primary_color_light="#fda4af",
            input_color="#1e293b",
            input_color_background="#ffffff",
            light_color="#e2e8f0",
            light_color_hover="#cbd5e1",
            text_color="#475569",
            accent_color="#1e293b",
            background_color="#ffffff",
            font_size=16,
            font_family="'Inter', sans-serif",
        )



from collections.abc import Callable, Coroutine

from fastapi import Depends, Header, Query
from sqlalchemy.sql import Select

from fief.repositories.base import BaseRepository, M

RawOrdering = list[str]
Ordering = list[tuple[list[str], bool]]
Pagination = tuple[int, int]
PaginatedObjects = tuple[list[M], int]
GetPaginatedObjects = Callable[
    [Select, Pagination, Ordering, BaseRepository[M]],
    Coroutine[None, None, PaginatedObjects[M]],
]


async def get_paginated_objects(
    statement: Select,
    pagination: Pagination,
    ordering: Ordering,
    repository: BaseRepository[M],
) -> PaginatedObjects[M]:
    limit, skip = pagination
    statement = repository.orderize(statement, ordering)
    return await repository.paginate(statement, limit, skip)


async def get_paginated_objects_noop(
    statement: Select,
    pagination: Pagination,
    ordering: Ordering,
    repository: BaseRepository[M],
) -> PaginatedObjects[M]:
    return ([], 0)


async def get_paginated_objects_getter(
    hx_target: str | None = Header(None, include_in_schema=False),
) -> GetPaginatedObjects[M]:
    if hx_target == "aside-content":
        return get_paginated_objects_noop
    return get_paginated_objects


async def get_pagination(
    limit: int = Query(10, ge=0), skip: int = Query(0, ge=0)
) -> Pagination:
    return min(limit, 100), skip


async def get_raw_ordering(ordering: str = Query(None)) -> RawOrdering:
    return ordering.split(",") if ordering else []


class OrderingGetter:
    def __init__(self, default: Ordering | None = None) -> None:
        self.default = default if default else []

    async def __call__(
        self,
        raw_ordering: RawOrdering = Depends(get_raw_ordering),
    ) -> Ordering:
        ordering_fields = []
        for field in raw_ordering:
            is_desc = False
            if field.startswith("-"):
                is_desc = True
                field = field[1:]
            ordering_fields.append((field.split("."), is_desc))

        if len(ordering_fields) == 0:
            return self.default

        return ordering_fields



import uuid

from sqlalchemy import select

from fief.models import Client
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class ClientRepository(BaseRepository[Client], UUIDRepositoryMixin[Client]):
    model = Client

    async def get_by_client_id(self, client_id: str) -> Client | None:
        statement = select(Client).where(Client.client_id == client_id)
        return await self.get_one_or_none(statement)

    async def get_by_client_id_and_tenant(
        self, client_id: str, tenant: uuid.UUID
    ) -> Client | None:
        statement = select(Client).where(
            Client.client_id == client_id, Client.tenant_id == tenant
        )
        return await self.get_one_or_none(statement)

    async def get_by_client_id_and_secret(
        self, client_id: str, client_secret: str
    ) -> Client | None:
        statement = select(Client).where(
            Client.client_id == client_id, Client.client_secret == client_secret
        )
        return await self.get_one_or_none(statement)

    async def count_by_tenant(self, tenant: uuid.UUID) -> int:
        statement = select(Client).where(Client.tenant_id == tenant)
        return await self._count(statement)



from datetime import datetime, timezone

from sqlalchemy import select

from fief.models import OAuthSession
from fief.repositories.base import BaseRepository, ExpiresAtMixin, UUIDRepositoryMixin


class OAuthSessionRepository(
    BaseRepository[OAuthSession],
    UUIDRepositoryMixin[OAuthSession],
    ExpiresAtMixin[OAuthSession],
):
    model = OAuthSession

    async def get_by_token(
        self, token: str, *, fresh: bool = True
    ) -> OAuthSession | None:
        statement = select(OAuthSession).where(OAuthSession.token == token)
        if fresh:
            statement = statement.where(
                OAuthSession.expires_at > datetime.now(timezone.utc)
            )
        return await self.get_one_or_none(statement)



from fastapi.responses import RedirectResponse
from starlette.background import BackgroundTask
from starlette.datastructures import URL


class HXRedirectResponse(RedirectResponse):
    def __init__(
        self,
        url: str | URL,
        status_code: int = 200,
        headers: dict[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(url, status_code, headers, background)
        self.headers["HX-Redirect"] = self.headers["location"]



from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import or_, select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import OAuthProvider, Tenant
from fief.repositories import OAuthProviderRepository


async def get_paginated_oauth_providers(
    query: str | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[OAuthProvider] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[OAuthProvider]:
    statement = select(OAuthProvider)

    if query is not None:
        statement = statement.where(
            or_(
                OAuthProvider.name.ilike(f"%{query}%"),
                OAuthProvider.provider.ilike(f"%{query}%"),
            )
        )

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_oauth_provider_by_id_or_404(
    id: UUID4,
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
) -> OAuthProvider:
    oauth_provider = await repository.get_by_id(id)

    if oauth_provider is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return oauth_provider


async def get_oauth_providers(
    tenant: Tenant = Depends(get_current_tenant),
) -> list[OAuthProvider]:
    return tenant.oauth_providers



from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import Tenant, Theme
from fief.repositories import (
    OAuthProviderRepository,
    ThemeRepository,
    UserFieldRepository,
)
from fief.services.theme_preview import ThemePreview


async def get_paginated_themes(
    query: str | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
    get_paginated_objects: GetPaginatedObjects[Theme] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Theme]:
    statement = select(Theme)

    if query is not None:
        statement = statement.where(Theme.name.ilike(f"%{query}%"))

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_theme_by_id_or_404(
    id: UUID4,
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
) -> Theme:
    theme = await repository.get_by_id(id)

    if theme is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return theme


async def get_theme_preview(
    oauth_provider_repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    user_field_repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
) -> ThemePreview:
    return ThemePreview(oauth_provider_repository, user_field_repository)


async def get_current_theme(
    tenant: Tenant = Depends(get_current_tenant),
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
) -> Theme:
    if tenant.theme_id is not None:
        theme = await repository.get_by_id(tenant.theme_id)
    else:
        theme = await repository.get_default()

    if theme is None:
        return Theme.build_default()

    return theme



from pydantic import UUID4

from fief.db.workspace import WorkspaceEngineManager, get_workspace_session
from fief.models import Client, Tenant, Workspace, WorkspaceUser
from fief.repositories import (
    ClientRepository,
    EmailTemplateRepository,
    TenantRepository,
    ThemeRepository,
    WorkspaceRepository,
    WorkspaceUserRepository,
)
from fief.schemas.workspace import WorkspaceCreate
from fief.services.email_template.initializer import EmailTemplateInitializer
from fief.services.localhost import is_localhost
from fief.services.main_workspace import get_main_fief_client, get_main_fief_workspace
from fief.services.theme import init_default_theme
from fief.services.workspace_db import (
    WorkspaceDatabase,
    WorkspaceDatabaseConnectionError,
)


class WorkspaceCreation:
    def __init__(
        self,
        workspace_repository: WorkspaceRepository,
        workspace_user_repository: WorkspaceUserRepository,
        workspace_db: WorkspaceDatabase,
        workspace_engine_manager: WorkspaceEngineManager,
    ) -> None:
        self.workspace_repository = workspace_repository
        self.workspace_user_repository = workspace_user_repository
        self.workspace_db = workspace_db
        self.workspace_engine_manager = workspace_engine_manager

    async def create(
        self,
        workspace_create: WorkspaceCreate,
        user_id: UUID4 | None = None,
        default_domain: str | None = None,
        default_client_id: str | None = None,
        default_client_secret: str | None = None,
        default_redirect_uris: list[str] | None = None,
        default_encryption_key: str | None = None,
    ) -> Workspace:
        workspace = Workspace(**workspace_create.dict())

        if default_domain is None:
            # Create workspace on available subdomain
            domain = await self.workspace_repository.get_available_subdomain(
                workspace.name
            )
            workspace.domain = domain
        else:
            workspace.domain = default_domain

        workspace = await self.workspace_repository.create(workspace)

        # Apply the database schema
        try:
            alembic_revision = self.workspace_db.migrate(
                workspace.get_database_connection_parameters(False),
                workspace.get_schema_name(),
            )
        except WorkspaceDatabaseConnectionError:
            await self.workspace_repository.delete(workspace)
            raise

        workspace.alembic_revision = alembic_revision
        await self.workspace_repository.update(workspace)

        # Link the user to this workspace
        if user_id is not None:
            workspace_user = WorkspaceUser(workspace_id=workspace.id, user_id=user_id)
            await self.workspace_user_repository.create(workspace_user)

        # Create a default tenant, client and email templates
        async with get_workspace_session(
            workspace, self.workspace_engine_manager
        ) as session:
            tenant_name = workspace.name
            tenant_slug = await TenantRepository(session).get_available_slug(
                tenant_name
            )
            tenant = Tenant(name=workspace.name, slug=tenant_slug, default=True)

            session.add(tenant)

            client = Client(
                name=f"{tenant.name}'s client",
                first_party=True,
                tenant=tenant,
                redirect_uris=default_redirect_uris
                if default_redirect_uris is not None
                else ["http://localhost:8000/docs/oauth2-redirect"],
            )

            if default_client_id is not None:
                client.client_id = default_client_id
            if default_client_secret is not None:
                client.client_secret = default_client_secret
            if default_encryption_key is not None:
                client.encrypt_jwk = default_encryption_key

            session.add(client)

            email_template_repository = EmailTemplateRepository(session)
            email_template_initializer = EmailTemplateInitializer(
                email_template_repository
            )
            await email_template_initializer.init_templates()

            theme_repository = ThemeRepository(session)
            await init_default_theme(theme_repository)

            await session.commit()

        # Allow a redirect URI on this workspace on the main Fief client
        await self._add_fief_redirect_uri(workspace)

        return workspace

    async def _add_fief_redirect_uri(self, workspace: Workspace):
        main_workspace = await get_main_fief_workspace()

        async with get_workspace_session(
            main_workspace, self.workspace_engine_manager
        ) as session:
            client_repository = ClientRepository(session)
            fief_client = await get_main_fief_client()

            is_localhost_domain = is_localhost(workspace.domain)
            redirect_uri = f"{'http' if is_localhost_domain else 'https'}://{workspace.domain}/admin/auth/callback"
            fief_client.redirect_uris = fief_client.redirect_uris + [redirect_uri]

            await client_repository.update(fief_client)



from fastapi import APIRouter, Depends, HTTPException, Response, status

from fief import schemas
from fief.dependencies.admin_authentication import is_authenticated_admin_api
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.role import get_paginated_roles, get_role_by_id_or_404
from fief.dependencies.tasks import get_send_task
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.errors import APIErrorCode
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Role, Workspace
from fief.repositories import PermissionRepository, RoleRepository
from fief.schemas.generics import PaginatedResults
from fief.services.webhooks.models import RoleCreated, RoleDeleted, RoleUpdated
from fief.tasks import SendTask, on_role_updated

router = APIRouter(dependencies=[Depends(is_authenticated_admin_api)])


@router.get(
    "/",
    name="roles:list",
    response_model=PaginatedResults[schemas.role.Role],
)
async def list_roles(
    paginated_roles: PaginatedObjects[Role] = Depends(get_paginated_roles),
) -> PaginatedResults[schemas.role.Role]:
    roles, count = paginated_roles
    return PaginatedResults(
        count=count,
        results=[schemas.role.Role.from_orm(role) for role in roles],
    )


@router.get("/{id:uuid}", name="roles:get", response_model=schemas.role.Role)
async def get_role(role: Role = Depends(get_role_by_id_or_404)) -> Role:
    return role


@router.post(
    "/",
    name="roles:create",
    response_model=schemas.role.Role,
    status_code=status.HTTP_201_CREATED,
)
async def create_role(
    role_create: schemas.role.RoleCreate,
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.role.Role:
    role = Role(**role_create.dict(exclude={"permissions"}))

    for permission_id in role_create.permissions:
        permission = await permission_repository.get_by_id(permission_id)
        if permission is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=APIErrorCode.ROLE_CREATE_NOT_EXISTING_PERMISSION,
            )
        else:
            role.permissions.append(permission)

    role = await repository.create(role)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, role)
    trigger_webhooks(RoleCreated, role, schemas.role.Role)

    return schemas.role.Role.from_orm(role)


@router.patch(
    "/{id:uuid}",
    name="roles:update",
    response_model=schemas.role.Role,
)
async def update_role(
    role_update: schemas.role.RoleUpdate,
    role: Role = Depends(get_role_by_id_or_404),
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    send_task: SendTask = Depends(get_send_task),
    workspace: Workspace = Depends(get_current_workspace),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
) -> schemas.role.Role:
    role_update_dict = role_update.dict(exclude_unset=True, exclude={"permissions"})
    for field, value in role_update_dict.items():
        setattr(role, field, value)

    old_permissions = {permission.id for permission in role.permissions}
    if role_update.permissions is not None:
        role.permissions = []
        for permission_id in role_update.permissions:
            permission = await permission_repository.get_by_id(permission_id)
            if permission is None:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=APIErrorCode.ROLE_UPDATE_NOT_EXISTING_PERMISSION,
                )
            else:
                role.permissions.append(permission)
    new_permissions = {permission.id for permission in role.permissions}

    await repository.update(role)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, role)
    trigger_webhooks(RoleUpdated, role, schemas.role.Role)

    added_permissions = new_permissions - old_permissions
    deleted_permissions = old_permissions - new_permissions
    send_task(
        on_role_updated,
        str(role.id),
        set(map(str, added_permissions)),
        set(map(str, deleted_permissions)),
        str(workspace.id),
    )

    return schemas.role.Role.from_orm(role)


@router.delete(
    "/{id:uuid}",
    name="roles:delete",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
)
async def delete_role(
    role: Role = Depends(get_role_by_id_or_404),
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    await repository.delete(role)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, role)
    trigger_webhooks(RoleDeleted, role, schemas.role.Role)



from fief.models import OAuthProvider, Tenant
from fief.schemas.auth import (
    AuthorizeError,
    AuthorizeRedirectError,
    LoginError,
    LogoutError,
    TokenError,
)
from fief.schemas.oauth import OAuthError


class AuthorizeException(Exception):
    def __init__(self, error: AuthorizeError, tenant: Tenant | None = None) -> None:
        self.error = error
        self.tenant = tenant


class AuthorizeRedirectException(Exception):
    def __init__(
        self,
        error: AuthorizeRedirectError,
        redirect_uri: str,
        response_mode: str,
        state: str | None,
        tenant: Tenant | None = None,
    ) -> None:
        self.error = error
        self.redirect_uri = redirect_uri
        self.response_mode = response_mode
        self.state = state
        self.tenant = tenant


class LoginException(Exception):
    def __init__(
        self,
        error: LoginError,
        oauth_providers: list[OAuthProvider] | None = None,
        tenant: Tenant | None = None,
        *,
        fatal: bool = False,
    ) -> None:
        self.error = error
        self.oauth_providers = oauth_providers
        self.tenant = tenant
        self.fatal = fatal


class OAuthException(Exception):
    def __init__(
        self,
        error: OAuthError,
        oauth_providers: list[OAuthProvider] | None = None,
        tenant: Tenant | None = None,
        *,
        fatal: bool = False,
    ) -> None:
        self.error = error
        self.oauth_providers = oauth_providers
        self.tenant = tenant
        self.fatal = fatal


class TokenRequestException(Exception):
    def __init__(self, error: TokenError) -> None:
        self.error = error


class LogoutException(Exception):
    def __init__(self, error: LogoutError, tenant: Tenant | None = None) -> None:
        self.error = error
        self.tenant = tenant



from typing import Any

from pydantic import BaseModel


class WebhookEventType:
    """Base webhook event type."""

    object: str
    event: str

    def __new__(cls):
        raise TypeError("WebhookEventType can't be instantiated.")  # noqa: TRY003

    @classmethod
    def key(cls) -> str:
        return f"{cls.object}.{cls.event}"

    @classmethod
    def description(cls) -> str | None:
        return cls.__doc__


class ClientCreated(WebhookEventType):
    """Occurs whenever a Client is created."""

    object = "client"
    event = "created"


class ClientUpdated(WebhookEventType):
    """Occurs whenever a Client is updated."""

    object = "client"
    event = "updated"


class ClientDeleted(WebhookEventType):
    """Occurs whenever a Client is deleted."""

    object = "client"
    event = "deleted"


class EmailTemplateUpdated(WebhookEventType):
    """Occurs whenever an Email Template is updated."""

    object = "email_template"
    event = "updated"


class OAuthProviderCreated(WebhookEventType):
    """Occurs whenever an OAuth Provider is created."""

    object = "oauth_provider"
    event = "created"


class OAuthProviderUpdated(WebhookEventType):
    """Occurs whenever an OAuth Provider is updated."""

    object = "oauth_provider"
    event = "updated"


class OAuthProviderDeleted(WebhookEventType):
    """Occurs whenever an OAuth Provider is deleted."""

    object = "oauth_provider"
    event = "deleted"


class PermissionCreated(WebhookEventType):
    """Occurs whenever a Permission is created."""

    object = "permission"
    event = "created"


class PermissionUpdated(WebhookEventType):
    """Occurs whenever a Permission is updated."""

    object = "permission"
    event = "updated"


class PermissionDeleted(WebhookEventType):
    """Occurs whenever a Permission is deleted."""

    object = "permission"
    event = "deleted"


class RoleCreated(WebhookEventType):
    """Occurs whenever a Role is created."""

    object = "role"
    event = "created"


class RoleUpdated(WebhookEventType):
    """Occurs whenever a Role is updated."""

    object = "role"
    event = "updated"


class RoleDeleted(WebhookEventType):
    """Occurs whenever a Role is deleted."""

    object = "role"
    event = "deleted"


class TenantCreated(WebhookEventType):
    """Occurs whenever a Tenant is created."""

    object = "tenant"
    event = "created"


class TenantUpdated(WebhookEventType):
    """Occurs whenever a Tenant is updated."""

    object = "tenant"
    event = "updated"


class TenantDeleted(WebhookEventType):
    """Occurs whenever a Tenant is deleted."""

    object = "tenant"
    event = "deleted"


class UserCreated(WebhookEventType):
    """Occurs whenever a User is created."""

    object = "user"
    event = "created"


class UserUpdated(WebhookEventType):
    """Occurs whenever a User is updated."""

    object = "user"
    event = "updated"


class UserDeleted(WebhookEventType):
    """Occurs whenever a User is deleted."""

    object = "user"
    event = "deleted"


class UserForgotPasswordRequested(WebhookEventType):
    """Occurs whenever a User requested to reset their password."""

    object = "user"
    event = "forgot_password_requested"


class UserPasswordReset(WebhookEventType):
    """Occurs whenever a User successfully reset their password."""

    object = "user"
    event = "password_reset"


class UserFieldCreated(WebhookEventType):
    """Occurs whenever a User Field is created."""

    object = "user_field"
    event = "created"


class UserFieldUpdated(WebhookEventType):
    """Occurs whenever a User Field is updated."""

    object = "user_field"
    event = "updated"


class UserFieldDeleted(WebhookEventType):
    """Occurs whenever a User Field is deleted."""

    object = "user_field"
    event = "deleted"


class UserPermissionCreated(WebhookEventType):
    """Occurs whenever a User Permission is created."""

    object = "user_permission"
    event = "created"


class UserPermissionDeleted(WebhookEventType):
    """Occurs whenever a User Permission is deleted."""

    object = "user_permission"
    event = "deleted"


class UserRoleCreated(WebhookEventType):
    """Occurs whenever a User Role is created."""

    object = "user_role"
    event = "created"


class UserRoleDeleted(WebhookEventType):
    """Occurs whenever a User Role is deleted."""

    object = "user_role"
    event = "deleted"


WEBHOOK_EVENTS: list[type[WebhookEventType]] = [
    ClientCreated,
    ClientUpdated,
    ClientDeleted,
    EmailTemplateUpdated,
    OAuthProviderCreated,
    OAuthProviderUpdated,
    OAuthProviderDeleted,
    PermissionCreated,
    PermissionUpdated,
    PermissionDeleted,
    RoleCreated,
    RoleUpdated,
    RoleDeleted,
    TenantCreated,
    TenantUpdated,
    TenantDeleted,
    UserCreated,
    UserUpdated,
    UserDeleted,
    UserForgotPasswordRequested,
    UserPasswordReset,
    UserFieldCreated,
    UserFieldUpdated,
    UserFieldDeleted,
    UserPermissionCreated,
    UserPermissionDeleted,
    UserRoleCreated,
    UserRoleDeleted,
]


class WebhookEvent(BaseModel):
    type: str
    data: dict[str, Any]



from wtforms import StringField, validators

from fief.forms import CSRFBaseForm


class PermissionCreateForm(CSRFBaseForm):
    name = StringField(
        "Name",
        validators=[validators.InputRequired()],
        render_kw={"placeholder": "Create Castle"},
    )
    codename = StringField(
        "Codename",
        validators=[validators.InputRequired()],
        render_kw={"placeholder": "castles:create"},
    )



from fief.schemas.generics import BaseModel, CreatedUpdatedAt, UUIDSchema
from fief.services.email_template.types import EmailTemplateType


class EmailTemplateUpdate(BaseModel):
    subject: str | None
    content: str | None


class EmailTemplateBase(UUIDSchema, CreatedUpdatedAt):
    type: EmailTemplateType
    subject: str
    content: str


class EmailTemplate(EmailTemplateBase):
    pass



from sqlalchemy import event
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine

from fief.db.types import DatabaseConnectionParameters
from fief.settings import settings


def create_engine(
    database_connection_parameters: DatabaseConnectionParameters,
) -> AsyncEngine:
    database_url, connect_args = database_connection_parameters
    engine = create_async_engine(
        database_url,
        connect_args=connect_args,
        echo=False,
        pool_recycle=settings.database_pool_recycle_seconds,
        pool_pre_ping=settings.database_pool_pre_ping,
    )
    dialect_name = engine.dialect.name

    # Special tweak for SQLite to better handle transaction
    # See: https://docs.sqlalchemy.org/en/14/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
    if dialect_name == "sqlite":

        @event.listens_for(engine.sync_engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            dbapi_connection.isolation_level = None

            # Enable SQLite foreign key support, which is not enabled by default
            # See: https://www.sqlite.org/foreignkeys.html#fk_enable
            dbapi_connection.execute("pragma foreign_keys=ON")

        @event.listens_for(engine.sync_engine, "begin")
        def do_begin(conn):
            # emit our own BEGIN
            conn.exec_driver_sql("BEGIN")

    return engine


def create_async_session_maker(engine: AsyncEngine):
    return async_sessionmaker(engine, expire_on_commit=False)


__all__ = [
    "create_engine",
    "create_async_session_maker",
]



from pydantic import UUID4

from fief.schemas.generics import BaseModel, CreatedUpdatedAt
from fief.schemas.permission import PermissionEmbedded
from fief.schemas.role import RoleEmbedded


class UserPermissionCreate(BaseModel):
    id: UUID4


class BaseUserPermission(CreatedUpdatedAt):
    user_id: UUID4
    permission_id: UUID4
    from_role_id: UUID4 | None


class UserPermission(BaseUserPermission):
    permission: PermissionEmbedded
    from_role: RoleEmbedded | None



import contextlib
from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING

import asyncpg.exceptions
import pymysql.err
from sqlalchemy import exc
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncEngine, AsyncSession

from fief.db.engine import create_engine
from fief.db.types import DatabaseConnectionParameters

if TYPE_CHECKING:
    from fief.models import Workspace


class WorkspaceEngineManager:
    def __init__(self) -> None:
        self.engines: dict[str, AsyncEngine] = {}

    def get_engine(
        self, database_connection_parameters: DatabaseConnectionParameters
    ) -> AsyncEngine:
        database_url, _ = database_connection_parameters
        key = str(database_url)
        try:
            return self.engines[key]
        except KeyError:
            engine = create_engine(database_connection_parameters)
            self.engines[key] = engine
            return engine

    async def close_all(self):
        for engine in self.engines.values():
            await engine.dispose()


@contextlib.asynccontextmanager
async def get_connection(
    engine: AsyncEngine, schema_name: str | None = None
) -> AsyncGenerator[AsyncConnection, None]:
    dialect_name = engine.dialect.name
    options = {}
    if dialect_name != "sqlite":
        options["schema_translate_map"] = {None: schema_name}
    try:
        async with engine.connect() as connection:
            yield await connection.execution_options(**options)
    except (
        asyncpg.exceptions.PostgresConnectionError,
        asyncpg.exceptions.InvalidAuthorizationSpecificationError,
        OSError,
    ) as e:
        raise ConnectionError from e
    except exc.OperationalError as e:
        # Catch MySQL connection error with code 2003
        if isinstance(e.orig, pymysql.err.OperationalError) and e.orig.args[0] == 2003:
            raise ConnectionError from e


@contextlib.asynccontextmanager
async def get_workspace_session(
    workspace: "Workspace",
    workspace_engine_manager: WorkspaceEngineManager,
) -> AsyncGenerator[AsyncSession, None]:
    engine = workspace_engine_manager.get_engine(
        workspace.get_database_connection_parameters()
    )
    async with get_connection(engine, workspace.get_schema_name()) as connection:
        async with AsyncSession(bind=connection, expire_on_commit=False) as session:
            yield session



from collections.abc import Mapping
from typing import Any

from fastapi.templating import Jinja2Templates
from jinja2 import pass_context, runtime
from starlette.background import BackgroundTask
from starlette.routing import Router

from fief.locale import get_translations
from fief.paths import TEMPLATES_DIRECTORY
from fief.services.oauth_provider import get_oauth_provider_branding


@pass_context
def get_column_macro(context: runtime.Context, column):
    column.renderer_macro = context[column.renderer_macro]
    return column


class LocaleJinja2Templates(Jinja2Templates):
    def _create_env(self, directory):
        env = super()._create_env(directory)
        env.add_extension("jinja2.ext.i18n")

        @pass_context
        def url_path_for(context: dict, name: str, **path_params: Any) -> str:
            router: Router = context["request"].scope["router"]
            return str(router.url_path_for(name, **path_params))

        env.globals["url_path_for"] = url_path_for
        env.globals["get_oauth_provider_branding"] = get_oauth_provider_branding
        env.filters["get_column_macro"] = get_column_macro
        env.install_gettext_translations(get_translations(), newstyle=True)  # type: ignore

        return env

    def TemplateResponse(
        self,
        name: str,
        context: dict,
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ):
        self.env.install_gettext_translations(get_translations(), newstyle=True)  # type: ignore
        return super().TemplateResponse(
            name, context, status_code, headers, media_type, background
        )


templates = LocaleJinja2Templates(directory=TEMPLATES_DIRECTORY)



from typing import TypeVar

from fastapi import Depends
from wtforms import EmailField, FormField, validators

from fief.dependencies.register import get_optional_registration_session
from fief.dependencies.user_field import get_registration_user_fields
from fief.forms import BaseForm, CSRFBaseForm, PasswordFieldForm, get_form_field
from fief.locale import gettext_lazy as _
from fief.models import RegistrationSession, RegistrationSessionFlow, UserField


class RegisterFormBase(CSRFBaseForm):
    email = EmailField(
        _("Email address"), validators=[validators.InputRequired(), validators.Email()]
    )
    fields: FormField


RF = TypeVar("RF", bound=RegisterFormBase)


async def get_register_form_class(
    registration_user_fields: list[UserField] = Depends(get_registration_user_fields),
    registration_session: RegistrationSession
    | None = Depends(get_optional_registration_session),
) -> type[RF]:
    class RegisterFormFields(BaseForm):
        pass

    for field in registration_user_fields:
        setattr(RegisterFormFields, field.slug, get_form_field(field))

    class RegisterForm(RegisterFormBase):
        fields = FormField(RegisterFormFields, separator=".")

    class RegisterPasswordForm(RegisterForm, PasswordFieldForm):
        pass

    if registration_session is None or (
        registration_session is not None
        and registration_session.flow == RegistrationSessionFlow.PASSWORD
    ):
        return RegisterPasswordForm

    return RegisterForm



from pydantic import UUID4
from sqlalchemy import select

from fief.models import WorkspaceUser
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class WorkspaceUserRepository(
    BaseRepository[WorkspaceUser], UUIDRepositoryMixin[WorkspaceUser]
):
    model = WorkspaceUser

    async def get_by_workspace_and_user(
        self, workspace_id: UUID4, user_id: UUID4
    ) -> WorkspaceUser | None:
        statement = select(WorkspaceUser).where(
            WorkspaceUser.workspace_id == workspace_id, WorkspaceUser.user_id == user_id
        )
        return await self.get_one_or_none(statement)



import functools

from fastapi import Depends

from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.tasks import get_send_task
from fief.models import Workspace
from fief.services.webhooks.trigger import TriggerWebhooks, trigger_webhooks
from fief.tasks import SendTask


async def get_trigger_webhooks(
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
) -> TriggerWebhooks:
    return functools.partial(
        trigger_webhooks, workspace_id=workspace.id, send_task=send_task
    )



from datetime import date, datetime
from typing import TYPE_CHECKING, Any

from pydantic import UUID4
from sqlalchemy import JSON, Boolean, Date, ForeignKey, Integer, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, TIMESTAMPAware, UUIDModel
from fief.models.user import User
from fief.models.user_field import UserField, UserFieldType

if TYPE_CHECKING:
    hybrid_property = property
else:
    from sqlalchemy.ext.hybrid import hybrid_property


class UserFieldValue(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "user_field_values"
    __table_args__ = (UniqueConstraint("user_field_id", "user_id"),)

    value_string: Mapped[str | None] = mapped_column(Text, nullable=True)
    value_integer: Mapped[int | None] = mapped_column(Integer, nullable=True)
    value_boolean: Mapped[bool | None] = mapped_column(Boolean, nullable=True)
    value_date: Mapped[date | None] = mapped_column(Date, nullable=True)
    value_datetime: Mapped[datetime | None] = mapped_column(
        TIMESTAMPAware(timezone=True), nullable=True
    )
    value_json: Mapped[Any | None] = mapped_column(JSON, nullable=True)

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    user: Mapped[User] = relationship("User", back_populates="user_field_values")

    user_field_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(UserField.id, ondelete="CASCADE"), nullable=False
    )
    user_field: Mapped[UserField] = relationship(
        "UserField", back_populates="user_field_values", lazy="selectin"
    )

    def _get_field_value(self) -> str:
        user_field_type = self.user_field.type
        if user_field_type in [UserFieldType.INTEGER]:
            return "value_integer"
        elif user_field_type in [UserFieldType.BOOLEAN]:
            return "value_boolean"
        elif user_field_type in [UserFieldType.DATE]:
            return "value_date"
        elif user_field_type in [UserFieldType.DATETIME]:
            return "value_datetime"
        elif user_field_type in [UserFieldType.ADDRESS]:
            return "value_json"
        else:
            return "value_string"

    @hybrid_property
    def value(self) -> Any:
        field_value = self._get_field_value()
        return getattr(self, field_value)

    @value.setter
    def value(self, value: Any):
        field_value = self._get_field_value()
        setattr(self, field_value, value)

    def get_slug_and_value(self, *, json_serializable: bool = False) -> tuple[str, Any]:
        value = self.value
        if json_serializable and (
            isinstance(value, date) or isinstance(value, datetime)
        ):
            value = value.isoformat()
        return self.user_field.slug, value



from fastapi import Depends, HTTPException, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import Webhook, WebhookLog
from fief.repositories import WebhookLogRepository, WebhookRepository


async def get_paginated_webhooks(
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[Webhook] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Webhook]:
    statement = select(Webhook)

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_webhook_by_id_or_404(
    id: UUID4,
    repository: WebhookRepository = Depends(
        get_workspace_repository(WebhookRepository)
    ),
) -> Webhook:
    webhook = await repository.get_by_id(id)

    if webhook is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return webhook


async def get_paginated_webhook_logs(
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter([(["created_at"], True)])),
    repository: WebhookLogRepository = Depends(
        get_workspace_repository(WebhookLogRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[WebhookLog] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[WebhookLog]:
    statement = select(WebhookLog).where(WebhookLog.webhook_id == webhook.id)

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_webhook_log_by_id_and_webhook_or_404(
    log_id: UUID4,
    webhook: Webhook = Depends(get_webhook_by_id_or_404),
    repository: WebhookLogRepository = Depends(
        get_workspace_repository(WebhookLogRepository)
    ),
) -> WebhookLog:
    webhook_log = await repository.get_by_id_and_webhook(log_id, webhook.id)

    if webhook_log is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return webhook_log



import json
import uuid
from datetime import datetime, timezone

from jwcrypto import jwk, jwt
from jwcrypto.common import JWException
from pydantic import UUID4

from fief.models import Client, User


class InvalidAccessToken(Exception):
    pass


def generate_access_token(
    key: jwk.JWK,
    host: str,
    client: Client,
    user: User,
    scope: list[str],
    permissions: list[str],
    lifetime_seconds: int,
) -> str:
    iat = int(datetime.now(timezone.utc).timestamp())
    exp = iat + lifetime_seconds

    claims = {
        "iss": host,
        "sub": str(user.id),
        "aud": [client.client_id],
        "exp": exp,
        "iat": iat,
        "azp": client.client_id,
        "scope": " ".join(scope),
        "permissions": permissions,
    }

    token = jwt.JWT(header={"alg": "RS256", "kid": key["kid"]}, claims=claims)
    token.make_signed_token(key)

    return token.serialize()


def read_access_token(key: jwk.JWK, token: str) -> UUID4:
    try:
        decoded_jwt = jwt.JWT(jwt=token, key=key)
        claims = json.loads(decoded_jwt.claims)
        user_id = claims["sub"]
    except (JWException, KeyError, ValueError) as e:
        raise InvalidAccessToken() from e
    else:
        return uuid.UUID(user_id)



from datetime import datetime
from typing import cast

from pydantic import UUID4
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.sqltypes import JSON, String

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import (
    GUID,
    CreatedUpdatedAt,
    ExpiresAt,
    TIMESTAMPAware,
    UUIDModel,
)
from fief.models.user import User


class AuthorizationCode(UUIDModel, CreatedUpdatedAt, ExpiresAt, WorkspaceBase):
    __tablename__ = "authorization_codes"

    code: Mapped[str] = mapped_column(
        String(length=255),
        nullable=False,
        index=True,
        unique=True,
    )
    c_hash: Mapped[str] = mapped_column(String(length=255), nullable=False)
    redirect_uri: Mapped[str] = mapped_column(String(length=2048), nullable=False)
    scope: Mapped[list[str]] = mapped_column(JSON, nullable=False, default=list)
    authenticated_at: Mapped[datetime] = mapped_column(
        TIMESTAMPAware(timezone=True), nullable=False
    )
    nonce: Mapped[str | None] = mapped_column(String(length=255), nullable=True)
    code_challenge: Mapped[str | None] = mapped_column(
        String(length=255), nullable=True
    )
    code_challenge_method: Mapped[str | None] = mapped_column(
        String(length=255), nullable=True
    )

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    user: Mapped[User] = relationship("User")

    client_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False
    )
    client: Mapped[Client] = relationship("Client", lazy="joined")

    def get_code_challenge_tuple(self) -> tuple[str, str] | None:
        if self.code_challenge is not None:
            return (self.code_challenge, cast(str, self.code_challenge_method))
        return None



import json
from typing import Any

from pydantic import UUID4
from sqlalchemy import Boolean, ForeignKey, Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.webhook import Webhook


class WebhookLog(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "webhook_logs"

    webhook_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Webhook.id, ondelete="CASCADE"), nullable=False
    )
    event: Mapped[str] = mapped_column(String(255), nullable=False)
    attempt: Mapped[int] = mapped_column(Integer, nullable=False)
    payload: Mapped[str] = mapped_column(Text, nullable=False)
    success: Mapped[bool] = mapped_column(Boolean, nullable=False)
    response: Mapped[str | None] = mapped_column(Text, nullable=True)
    error_type: Mapped[str | None] = mapped_column(String(255), nullable=True)
    error_message: Mapped[str | None] = mapped_column(Text, nullable=True)

    webhook: Mapped[Webhook] = relationship("Webhook")

    @property
    def payload_dict(self) -> dict[str, Any]:
        return json.loads(self.payload)



from sqlalchemy import String
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class Permission(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "permissions"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    codename: Mapped[str] = mapped_column(
        String(length=255), nullable=False, unique=True
    )

    def __repr__(self) -> str:
        return f"Permission(id={self.id}, name={self.name}, codename={self.codename})"
